import logging
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor
from copy import deepcopy
from typing import Any, Dict, Optional

import click
from tqdm import tqdm

from panoramic.cli.companies.client import CompaniesClient
from panoramic.cli.config.analytics import is_enabled as analytics_module_is_enabled
from panoramic.cli.config.storage import update_config
from panoramic.cli.context import get_company_slug
from panoramic.cli.controller import reconcile
from panoramic.cli.diff import echo_diff
from panoramic.cli.errors import (
    InvalidDatasetException,
    InvalidModelException,
    JoinException,
    ValidationError,
    ValidationErrorSeverity,
)
from panoramic.cli.file_utils import write_yaml
from panoramic.cli.identifier_generator import IdentifierGenerator
from panoramic.cli.join_detector import JoinDetector
from panoramic.cli.local import get_state as get_local_state
from panoramic.cli.local.executor import LocalExecutor
from panoramic.cli.local.writer import FileWriter
from panoramic.cli.model_mapper import map_columns_to_model
from panoramic.cli.pano_model import PanoField, PanoModel, PanoModelJoin
from panoramic.cli.paths import Paths
from panoramic.cli.physical_data_source.client import PhysicalDataSourceClient
from panoramic.cli.print import echo_error, echo_errors, echo_info, echo_warnings
from panoramic.cli.refresh import Refresher
from panoramic.cli.remote import get_state as get_remote_state
from panoramic.cli.remote.executor import RemoteExecutor
from panoramic.cli.scan import Scanner, scan_fields_for_errors
from panoramic.cli.state import Action, ActionList
from panoramic.cli.validate import (
    validate_config,
    validate_context,
    validate_local_state,
    validate_missing_files,
    validate_orphaned_files,
)

logger = logging.getLogger(__name__)


def configure_anonymous_analytics():
    if not analytics_module_is_enabled():
        return

    analytics_opt_out = click.confirm(
        "\nDo you want to provide usage metrics to Panoramic?\n"
        "No passwords,credentials, or personal information will be shared.",
        default=True,
    )
    update_config('analytics', {'enabled': analytics_opt_out})


def configure():
    """Global configuration for CLI."""
    client_id = click.prompt('Enter your client id', type=str)
    client_secret = click.prompt('Enter your client secret', hide_input=True, type=str)
    update_config('auth', {'client_id': client_id, 'client_secret': client_secret})

    configure_anonymous_analytics()


def initialize():
    """Initialize context."""
    client = CompaniesClient()

    try:
        companies = client.get_companies()
    except Exception:
        logger.debug('Failed to fetch available companies', exc_info=True)
        companies = []

    base_text = 'Enter your company slug'
    if len(companies) == 0:
        prompt_text = base_text
    elif len(companies) > 3:
        prompt_text = f'{base_text} (Available - {{{",".join(list(companies)[:3])}}},...)'
    else:
        prompt_text = f'{base_text} (Available - {{{",".join(companies)}}})'

    company_slug = click.prompt(prompt_text, type=str, default=next(iter(companies), None))
    write_yaml(Paths.context_file(), {'company_slug': company_slug, 'api_version': 'v1'})


def list_connections():
    """List available data connections for company from context."""
    client = PhysicalDataSourceClient()

    sources = client.get_sources(get_company_slug())
    if len(sources) == 0:
        echo_error('No data connections have been found')
    else:
        for source in sources:
            echo_info(source['source_name'])


def list_companies():
    """List available companies for user."""
    client = CompaniesClient()
    companies = client.get_companies()
    if len(companies) == 0:
        echo_error('No companies have been found')
    else:
        for company in companies:
            echo_info(company)


def validate() -> bool:
    """Check local files against schema."""
    errors = []

    try:
        validate_config()
    except ValidationError as e:
        errors.append(e)

    try:
        validate_context()
    except ValidationError as e:
        errors.append(e)

    errors.extend(validate_local_state())

    errors_by_severity = defaultdict(list)
    for error in errors:
        errors_by_severity[error.severity].append(error)

    if len(errors_by_severity[ValidationErrorSeverity.WARNING]) > 0:
        echo_warnings(errors_by_severity[ValidationErrorSeverity.WARNING])
        echo_info('')

    if len(errors_by_severity[ValidationErrorSeverity.ERROR]) > 0:
        echo_errors(errors_by_severity[ValidationErrorSeverity.ERROR])
        return False

    echo_info("Success: All files are valid.")
    return True


def scan(source_id: str, table_filter: Optional[str], parallel: int = 1, generate_identifiers: bool = False):
    """Scan all metadata for given source and filter."""
    company_slug = get_company_slug()
    scanner = Scanner(company_slug, source_id)
    scanner.fetch_token()

    tables = list(scanner.scan_tables(table_filter=table_filter))

    if len(tables) == 0:
        echo_info('No tables have been found')
        return

    refresher = Refresher(company_slug, source_id)
    refresher.fetch_token()

    if generate_identifiers:
        id_generator = IdentifierGenerator(company_slug, source_id)
        id_generator.fetch_token()

    writer = FileWriter()

    progress_bar = tqdm(total=len(tables))

    def _process_table(table: Dict[str, Any]):
        # drop source name from table name
        table_name = table['data_source'].split('.', 1)[1]

        try:
            refresher.refresh_table(table_name)
            if generate_identifiers:
                identifiers = id_generator.generate(table_name)
            else:
                identifiers = []

            columns = list(scanner.scan_columns(table_filter=table_name))
            for model in map_columns_to_model(columns):
                model.identifiers = identifiers
                writer.write_scanned_model(model)
                progress_bar.write(f'Discovered model {model.model_name}')
        except Exception:
            error_msg = f'Metadata could not be scanned for table {table_name}'
            progress_bar.write(f'Error: {error_msg}')
            logger.debug(error_msg, exc_info=True)
            # Create an empty model file even when scan fails
            writer.write_empty_model(table['model_name'])
        finally:
            progress_bar.update()

    executor = ThreadPoolExecutor(max_workers=parallel)
    for _ in executor.map(_process_table, tables):
        pass

    progress_bar.write(f'Scanned {progress_bar.total} tables')


def pull(yes: bool = False, target_dataset: Optional[str] = None, diff: bool = False):
    """Pull models and data sources from remote."""
    company_slug = get_company_slug()
    echo_info('Loading local state...')
    local_state = get_local_state(target_dataset=target_dataset)
    echo_info('Fetching remote state...')
    remote_state = get_remote_state(company_slug, target_dataset=target_dataset)
    echo_info('Reconciling actions...')
    actions = reconcile(current_state=local_state, desired_state=remote_state)
    echo_info('')  # Add a blank row

    if len(actions.actions) == 0:
        if remote_state.is_empty:
            echo_info('No models have been published')
        else:
            echo_info('Local state is up to date')
        return

    echo_diff(actions)

    if diff:
        # User only wants diff print
        return

    if not yes and not click.confirm('Do you want to proceed?'):
        # User decided not to pull
        return

    executor = LocalExecutor()
    with tqdm(actions.actions) as bar:
        for action in bar:
            try:
                executor.execute(action)
            except Exception:
                bar.write(f'Error: Failed to execute action {action.description}')
        bar.write(f'Pulled {executor.success_count}/{executor.total_count} models, fields and datasets')


def push(yes: bool = False, target_dataset: Optional[str] = None, diff: bool = False):
    """Push models and data sources to remote."""
    company_slug = get_company_slug()
    echo_info('Loading local state...')
    local_state = get_local_state(target_dataset=target_dataset)
    echo_info('Fetching remote state...')
    remote_state = get_remote_state(company_slug, target_dataset=target_dataset)
    echo_info('Resolving state...')
    actions = reconcile(remote_state, local_state)
    echo_info('')  # Add a blank row

    if len(actions.actions) == 0:
        if local_state.is_empty:
            echo_info('No models to publish')
        else:
            echo_info('Remote state is up to date')
        return

    echo_diff(actions)

    if diff:
        # User only wants diff print
        return

    if not yes and not click.confirm('Do you want to proceed?'):
        # User decided not to push
        return

    executor = RemoteExecutor(company_slug)
    with tqdm(actions.actions) as bar:
        for action in bar:
            try:
                executor.execute(action)
            except (InvalidModelException, InvalidDatasetException) as e:
                messages_concat = '\n  '.join(e.messages)
                bar.write(f'Error: Failed to execute action {action.description}:\n  {messages_concat}')
            except Exception as e:
                bar.write(f'Error: Failed to execute action {action.description}:\n  {str(e)}')
        bar.write(f'Updated {executor.success_count}/{executor.total_count} models, fields and datasets')


def detect_joins(target_dataset: Optional[str] = None, diff: bool = False, overwrite: bool = False, yes: bool = False):
    company_slug = get_company_slug()
    echo_info('Loading local state...')
    local_state = get_local_state(target_dataset=target_dataset)

    if local_state.is_empty:
        echo_info('No datasets to detect joins on')
        return

    join_detector = JoinDetector(company_slug=company_slug)
    join_detector.fetch_token()

    models_by_virtual_data_source: Dict[Optional[str], Dict[str, PanoModel]] = defaultdict(dict)
    for model in local_state.models:
        # Prepare a mapping for a quick access when reconciling necessary changes later
        models_by_virtual_data_source[model.virtual_data_source][model.model_name] = model

    action_list: ActionList[PanoModel] = ActionList()

    with tqdm(list(local_state.data_sources)) as bar:
        for dataset in bar:
            try:
                bar.write(f'Detecting joins for dataset {dataset.dataset_slug}')
                joins_by_model = join_detector.detect(dataset.dataset_slug)

                for model_name, joins in joins_by_model.items():
                    if not joins:
                        bar.write(f'No joins detected for {model_name} under dataset {dataset.dataset_slug}')
                        continue

                    bar.write(f'Detected {len(joins)} joins for {model_name} under dataset {dataset.dataset_slug}')

                    detected_join_objects = [PanoModelJoin.from_dict(join_dict) for join_dict in joins]
                    current_model = models_by_virtual_data_source[dataset.dataset_slug][model_name]
                    desired_model = deepcopy(current_model)

                    if overwrite:
                        desired_model.joins = detected_join_objects
                    else:
                        for detected_join in detected_join_objects:
                            # Only append joins that are not already defined
                            if detected_join not in current_model.joins:
                                desired_model.joins.append(detected_join)

                    action_list.actions.append(Action(current=current_model, desired=desired_model))

            except JoinException as join_exception:
                bar.write(f'Error: {str(join_exception)}')
                logger.debug(str(join_exception), exc_info=True)
            except Exception:
                error_msg = f'An unexpected error occured when detecting joins for {dataset.dataset_slug}'
                bar.write(f'Error: {error_msg}')
                logger.debug(error_msg, exc_info=True)
            finally:
                bar.update()

    if action_list.is_empty:
        echo_info('No joins detected')
        return

    echo_diff(action_list)
    if diff:
        # User decided to see the diff only
        return

    if not yes and not click.confirm('Do you want to proceed?'):
        # User decided not to update local models based on join suggestions
        return

    echo_info('Updating local state...')

    executor = LocalExecutor()
    for action in action_list.actions:
        try:
            executor.execute(action)
        except Exception:
            echo_error(f'Error: Failed to execute action {action.description}')
        echo_info(f'Updated {executor.success_count}/{executor.total_count} models')


def delete_orphaned_fields(target_dataset: Optional[str] = None, yes: bool = False):
    """Delete orphaned field files."""
    echo_info('Loading local state...')
    state = get_local_state(target_dataset=target_dataset)

    action_list: ActionList[PanoField] = ActionList()

    for dataset, (fields, models) in state.get_objects_by_package().items():
        fields_by_slug = {f.slug: f for f in fields}
        for idx, error in enumerate(validate_orphaned_files(fields, models, package_name=dataset)):
            if idx == 0:
                echo_info(f'\nFields without calculation or reference in a model in dataset {dataset}:')
            echo_info(f'  {error.field_slug}')
            # Add deletion action
            action_list.add_action(Action(current=fields_by_slug[error.field_slug], desired=None))

    if action_list.is_empty:
        echo_info('No issues found')
        return

    echo_info('')
    if not yes and not click.confirm('You will not be able to query these fields. Do you want to remove them?'):
        # User decided not to fix issues
        return

    echo_info('Updating local state...')

    executor = LocalExecutor()
    for action in action_list.actions:
        try:
            executor.execute(action)
        except Exception:
            echo_error(f'Error: Failed to execute action {action.description}')
    echo_info(f'Updated {executor.success_count}/{executor.total_count} fields')


def scaffold_missing_fields(target_dataset: Optional[str] = None, yes: bool = False):
    """Scaffold missing field files."""
    echo_info('Loading local state...')
    state = get_local_state(target_dataset=target_dataset)

    errors = []

    for dataset, (fields, models) in state.get_objects_by_package().items():
        for idx, error in enumerate(validate_missing_files(fields, models, package_name=dataset)):
            if idx == 0:
                echo_info(f'\nFields referenced in models without definition in dataset {dataset}:')
            echo_info(f'  {error.field_slug}')
            errors.append(error)

    if len(errors) == 0:
        echo_info('No issues found')
        return

    echo_info('')
    if not yes and not click.confirm(
        'You will not be able to query these fields until you define them. Do you want to do that now?'
    ):
        # User decided not to fix issues
        return

    # TODO: improve messaging?
    echo_info('Scanning fields...')
    fields = scan_fields_for_errors(errors)
    action_list = ActionList(actions=[Action(desired=field) for field in fields])

    echo_info('Updating local state...')

    executor = LocalExecutor()
    for action in action_list.actions:
        try:
            executor.execute(action)
        except Exception:
            echo_error(f'Error: Failed to execute action {action.description}')
    echo_info(f'Updated {executor.success_count}/{executor.total_count} fields')



import itertools
from typing import Iterable, Optional

import requests
from requests.exceptions import RequestException

from panoramic.cli.errors import (
    CompanyNotFoundException,
    DatasetReadException,
    FieldReadException,
    ModelReadException,
)
from panoramic.cli.field.client import FieldClient
from panoramic.cli.field_mapper import map_field_from_remote
from panoramic.cli.model import ModelClient
from panoramic.cli.model_mapper import (
    map_data_source_from_remote,
    map_model_from_remote,
)
from panoramic.cli.pano_model import PanoField, PanoModel, PanoVirtualDataSource
from panoramic.cli.state import VirtualState
from panoramic.cli.virtual_data_source import VirtualDataSourceClient


def get_data_sources(company_slug: str, *, limit: int = 100) -> Iterable[PanoVirtualDataSource]:
    """Get virtual data sources from remote."""
    client = VirtualDataSourceClient()
    offset = 0
    while True:
        try:
            sources = client.get_all_virtual_data_sources(company_slug, offset=offset, limit=limit)
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise CompanyNotFoundException(company_slug).extract_request_id(e)
            raise DatasetReadException(company_slug).extract_request_id(e)

        yield from (map_data_source_from_remote(s) for s in sources)
        if len(sources) < limit:
            # last page
            break

        offset += limit


def get_models(data_source: str, company_slug: str, *, limit: int = 100) -> Iterable[PanoModel]:
    """Get all models from remote."""
    client = ModelClient()
    offset = 0
    while True:
        try:
            models = client.get_models(data_source, company_slug, offset=offset, limit=limit)
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise CompanyNotFoundException(company_slug).extract_request_id(e)
            raise ModelReadException(company_slug, data_source).extract_request_id(e)

        yield from (map_model_from_remote(m) for m in models)
        if len(models) < limit:
            # last page
            break

        offset += limit


def get_fields(company_slug: str, data_source: Optional[str] = None, *, limit: int = 100) -> Iterable[PanoField]:
    """Get all models from remote."""
    client = FieldClient()
    offset = 0
    while True:
        try:
            fields = client.get_fields(company_slug=company_slug, data_source=data_source, offset=offset, limit=limit)
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise CompanyNotFoundException(company_slug).extract_request_id(e)
            raise FieldReadException(company_slug, data_source).extract_request_id(e)

        yield from (map_field_from_remote(field) for field in fields)
        if len(fields) < limit:
            # last page
            break

        offset += limit


def get_state(company_slug: str, target_dataset: Optional[str] = None) -> VirtualState:
    """Build a representation of what VDS and models are on remote."""
    data_source_iterator = get_data_sources(company_slug)
    data_sources = (
        # Filter only targeted data source if the argument is set
        [vds for vds in data_source_iterator if vds.dataset_slug == target_dataset]
        if target_dataset
        else list(data_source_iterator)
    )
    models = list(
        itertools.chain.from_iterable(get_models(source.dataset_slug, company_slug) for source in data_sources)
    )
    fields = list(get_fields(company_slug=company_slug, data_source=target_dataset))
    return VirtualState(data_sources=data_sources, models=models, fields=fields)



import pytest
import responses

from panoramic.cli.virtual_data_source.client import (
    VirtualDataSource,
    VirtualDataSourceClient,
)


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_upsert_virtual_data_source():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.PUT,
        'https://diesel/virtual?company_slug=test-company',
        json={'display_name': 'test', 'slug': 'bug'},
    )

    client = VirtualDataSourceClient(
        base_url='https://diesel/virtual/', client_id='client-id', client_secret='client-secret'
    )
    client.upsert_virtual_data_source('test-company', VirtualDataSource(display_name='test', slug='bug'))


@responses.activate
def test_get_all_virtual_data_sources():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})

    fake_source = VirtualDataSource(display_name='virtual_source', slug='test-source')
    responses.add(
        responses.GET,
        'https://diesel/virtual?company_slug=test-company&offset=200&limit=100',
        json={'data': [fake_source.to_dict()]},
    )

    client = VirtualDataSourceClient(
        base_url='https://diesel/virtual/', client_id='client-id', client_secret='client-secret'
    )

    assert client.get_all_virtual_data_sources('test-company', offset=200, limit=100) == [fake_source]


@responses.activate
def test_get_virtual_data_source():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})

    fake_source = VirtualDataSource(display_name='virtual_source', slug='test-source')
    responses.add(
        responses.GET,
        'https://diesel/virtual/test-source?company_slug=test-company',
        json={'data': fake_source.to_dict()},
    )

    client = VirtualDataSourceClient(
        base_url='https://diesel/virtual/', client_id='client-id', client_secret='client-secret'
    )
    remote_source = client.get_virtual_data_source('test-company', fake_source.slug)

    assert remote_source == fake_source


@responses.activate
def test_delete_virtual_data_source():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(responses.DELETE, 'https://diesel/virtual/test-source?company_slug=test-company')

    client = VirtualDataSourceClient(
        base_url='https://diesel/virtual/', client_id='client-id', client_secret='client-secret'
    )
    client.delete_virtual_data_source('test-company', 'test-source')



import os
from pathlib import Path
from unittest.mock import Mock

from panoramic.cli.local.reader import FilePackage, FileReader
from panoramic.cli.paths import FileExtension, Paths, PresetFileName, SystemDirectory


def test_reader_get_packages(tmp_path: Path):
    # scanned directory
    scanned_dir = tmp_path / SystemDirectory.SCANNED.value
    scanned_dir.mkdir()
    (scanned_dir / PresetFileName.DATASET_YAML.value).touch()

    # dataset with one model
    ds1_dir = tmp_path / 'dataset1'
    ds1_dir.mkdir()

    ds1_file = ds1_dir / PresetFileName.DATASET_YAML.value
    ds1_file.touch()

    model_file = ds1_dir / f'test_model{FileExtension.MODEL_YAML.value}'
    model_file.touch()

    ds1_fields_dir = Paths.fields_dir(ds1_dir)
    ds1_fields_dir.mkdir()
    field_file = ds1_fields_dir / f'test_field{FileExtension.FIELD_YAML.value}'
    field_file.touch()

    # empty dataset
    ds2_dir = tmp_path / 'dataset2'
    ds2_dir.mkdir()

    packages = list(FileReader(cwd=tmp_path).get_packages())
    expected = [
        FilePackage(name='dataset1', data_source_file=ds1_file, model_files=[model_file], field_files=[field_file])
    ]
    assert packages == expected


def test_file_package_read_data_source(tmp_path):
    ds_file = tmp_path / PresetFileName.DATASET_YAML.value
    with ds_file.open('w') as f:
        f.write('dataset_slug: test-dataset')

    package = FilePackage(name='dataset1', data_source_file=ds_file, model_files=[], field_files=[])

    assert package.read_data_source() == {'dataset_slug': 'test-dataset'}


def test_file_package_read_models(tmp_path):
    model_file = tmp_path / f'test_model{FileExtension.MODEL_YAML.value}'
    with model_file.open('w') as f:
        f.write('model_name: test-model')

    package = FilePackage(name='dataset1', data_source_file=Mock(), model_files=[model_file], field_files=[])

    assert list(package.read_models()) == [({'model_name': 'test-model'}, model_file)]


def test_file_package_read_fields(tmp_path):
    field_file = Paths.fields_dir(tmp_path) / f'test_field{FileExtension.FIELD_YAML.value}'
    os.makedirs(os.path.dirname(field_file), exist_ok=True)

    with field_file.open('w') as f:
        f.write('slug: field_slug')

    package = FilePackage(name='dataset1', data_source_file=Mock(), model_files=[], field_files=[field_file])

    assert list(package.read_fields()) == [({'slug': 'field_slug'}, field_file)]



import pytest
import responses

from panoramic.cli.join.client import TERMINAL_STATES, JobState, JoinClient


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_create_join_detection_job():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.POST,
        'https://joins/test-dataset?company_slug=test-company',
        json={'data': {'job_id': 'test-job-id'}},
    )

    client = JoinClient(base_url='https://joins/', client_id='client-id', client_secret='client-secret')

    assert client.create_join_detection_job('test-company', 'test-dataset') == 'test-job-id'


@responses.activate
def test_get_job_status():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.GET, 'https://joins/job/test-job-id?company_slug=test-company', json={'data': {'status': 'RUNNING'}}
    )

    client = JoinClient(base_url='https://joins/', client_id='client-id', client_secret='client-secret')

    assert client.get_job_status(company_slug='test-company', job_id='test-job-id') == JobState.RUNNING


#
@responses.activate
def test_get_job_results():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.GET,
        'https://joins/job/test-job-id?company_slug=test-company',
        json={'data': {'job_id': 'test-job-id', 'status': 'COMPLETED', 'joins': {}}},
    )

    client = JoinClient(base_url='https://joins/', client_id='client-id', client_secret='client-secret')

    assert client.get_job_results(company_slug='test-company', job_id='test-job-id') == {
        'job_id': 'test-job-id',
        'status': 'COMPLETED',
        'joins': {},
    }


@responses.activate
@pytest.mark.parametrize('final_state', TERMINAL_STATES)
def test_wait_for_terminal_state(final_state):
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.GET, 'https://joins/job/test-job-id?company_slug=test-company', json={'data': {'status': 'RUNNING'}}
    )
    responses.add(
        responses.GET,
        'https://joins/job/test-job-id?company_slug=test-company',
        json={'data': {'status': final_state.value}},
    )

    client = JoinClient(base_url='https://joins/', client_id='client-id', client_secret='client-secret')

    assert client.wait_for_terminal_state(company_slug='test-company', job_id='test-job-id') == final_state



from panoramic.cli.__version__ import __version__


class VersionedClient:

    """Adds version headers to requests session."""

    def __init__(self):
        assert hasattr(self, 'session'), 'Client needs to have requests.Session'
        self.session.headers['User-Agent'] = f'pano-cli/{__version__}'  # type: ignore



from abc import ABC

from panoramic.cli.state import Action


class Executor(ABC):

    """Base executor class."""

    success_count: int = 0
    failure_count: int = 0
    total_count: int = 0

    def execute(self, action: Action):
        try:
            self._execute(action)
            self.success_count += 1
        except Exception:
            self.failure_count += 1
            raise
        finally:
            self.total_count += 1

    def _execute(self, action: Action):
        raise NotImplementedError('execute not implemented in base class')



import responses

from panoramic.cli.supported_version import URL, is_version_supported


@responses.activate
def test_not_found():
    responses.add(responses.GET, URL, json={'error': 'not found'}, status=404)
    assert is_version_supported("0.1.0") is False


@responses.activate
def test_empty_answer():
    responses.add(responses.GET, URL, json='')
    assert is_version_supported("0.1.0") is False


@responses.activate
def test_invalid_json():
    responses.add(responses.GET, URL, json='{"": }')
    assert is_version_supported("0.1.0") is False


@responses.activate
def test_success():
    responses.add(responses.GET, URL, json={"minimum_supported_version": "0.1.0"})
    assert is_version_supported("0.1.0") is True


@responses.activate
def test_success_with_cli_type():
    responses.add(responses.GET, URL, json={"minimum_supported_version": "0.1.0-python"})
    assert is_version_supported("0.1.0") is True


@responses.activate
def test_failure():
    responses.add(responses.GET, URL, json={"minimum_supported_version": "0.1.1"})
    assert is_version_supported("0.1.0") is False



import logging
import time
from enum import Enum
from typing import Any, Dict, Optional
from urllib.parse import urljoin

from panoramic.auth import OAuth2Client
from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.config.identifier import get_base_url
from panoramic.cli.errors import TimeoutException

logger = logging.getLogger(__name__)


class JobState(Enum):

    RUNNING = 'RUNNING'
    COMPLETED = 'COMPLETED'
    FAILED = 'FAILED'


TERMINAL_STATES = {JobState.COMPLETED, JobState.FAILED}


class IdentifierClient(OAuth2Client):

    """Identifier Parser HTTP API client."""

    base_url: str

    def __init__(
        self, base_url: Optional[str] = None, client_id: Optional[str] = None, client_secret: Optional[str] = None
    ):
        client_id = client_id if client_id is not None else get_client_id()
        client_secret = client_secret if client_secret is not None else get_client_secret()

        self.base_url = base_url if base_url is not None else get_base_url()

        super().__init__(client_id, client_secret)

    def create_identifier_job(self, company_slug: str, source_id: str, table_name: str) -> str:
        """Starts async "id parsing" job and return job id."""
        url = urljoin(self.base_url, f'{source_id}')
        params = {'company_slug': company_slug, 'table_name': table_name}
        logger.debug(f'Triggering a job for source {source_id} and table name: {table_name}')
        response = self.session.post(url, params=params, timeout=5)
        response.raise_for_status()
        return response.json()['data']['job_id']

    def get_job_status(self, job_id: str) -> JobState:
        """Get status of an async job."""
        url = urljoin(self.base_url, f'job/{job_id}')
        logger.debug(f'Getting job status for job {job_id}')
        response = self.session.get(url, timeout=5)
        response.raise_for_status()
        return JobState(response.json()['data']['status'])

    def get_job_results(self, job_id: str) -> Dict[str, Any]:
        """Get results of an async job."""
        url = urljoin(self.base_url, f'job/{job_id}')
        logger.debug(f'Getting job results for job {job_id}')
        response = self.session.get(url, timeout=5)
        response.raise_for_status()
        return response.json()['data']

    def wait_for_terminal_state(self, job_id: str, timeout: int = 60) -> JobState:
        """Wait for job to reach terminal state."""
        tick_time = 1
        while True:
            logger.debug(f'Getting status for job with id {job_id}')
            status = self.get_job_status(job_id)
            logger.debug(f'Got status {status} for job with id {job_id}')
            if status in TERMINAL_STATES:
                return status
            if timeout <= 0:
                raise TimeoutException(f'Timed out waiting for job {job_id} to complete')
            time.sleep(tick_time)
            timeout -= tick_time



import logging
from typing import Any, Dict, List

import requests
from requests import RequestException

from panoramic.cli.errors import DatasetNotFoundException, JoinException
from panoramic.cli.join import JobState, JoinClient

logger = logging.getLogger(__name__)


class JoinDetector:
    def __init__(self, company_slug: str, client: JoinClient = None):
        self.company_slug = company_slug

        if client is None:
            client = JoinClient()

        self.client = client

    def fetch_token(self):
        self.client.fetch_token()

    def detect(self, dataset_id: str, timeout: int = 60) -> Dict[str, List[Dict[str, Any]]]:
        """Detect possible joins between models under a dataset."""
        logger.debug(f'Starting join detection job for dataset {dataset_id}')
        try:
            job_id = self.client.create_join_detection_job(self.company_slug, dataset_id)
            logger.debug(f'Join detection job with id {job_id} started for dataset {dataset_id}')
        except RequestException as e:
            if e.response is not None and e.response.status_code == requests.codes.not_found:
                raise DatasetNotFoundException(dataset_id).extract_request_id(e)
            raise JoinException(dataset_id).extract_request_id(e)

        try:
            state = self.client.wait_for_terminal_state(company_slug=self.company_slug, job_id=job_id, timeout=timeout)
            if state != JobState.COMPLETED:
                raise JoinException(dataset_id)

            logger.debug(f'Join detection job with id {job_id} completed for dataset {dataset_id}')
            return self.client.get_job_results(company_slug=self.company_slug, job_id=job_id)['joins']
        except RequestException as e:
            raise JoinException(dataset_id).extract_request_id(e)



from unittest.mock import Mock, call, patch, sentinel

import pytest

from panoramic.cli.command import (
    delete_orphaned_fields,
    list_connections,
    push,
    scaffold_missing_fields,
    scan,
)
from panoramic.cli.errors import (
    InvalidDatasetException,
    InvalidModelException,
    OrphanFieldFileError,
)
from panoramic.cli.local.executor import LocalExecutor
from panoramic.cli.remote.executor import RemoteExecutor
from panoramic.cli.state import Action


@pytest.fixture(autouse=True)
def mock_get_company_slug():
    with patch('panoramic.cli.command.get_company_slug', return_value='test-company') as mock_get_company_slug:
        yield mock_get_company_slug


@pytest.fixture
def mock_refresher():
    with patch('panoramic.cli.command.Refresher') as mock_refresher:
        yield mock_refresher()


@pytest.fixture
def mock_scanner():
    with patch('panoramic.cli.command.Scanner') as mock_scanner:
        yield mock_scanner()


@pytest.fixture
def mock_id_generator():
    with patch('panoramic.cli.command.IdentifierGenerator') as mock_id_generator:
        yield mock_id_generator()


@pytest.fixture
def mock_writer():
    with patch('panoramic.cli.command.FileWriter', autospec=True) as mock_writer:
        yield mock_writer()


def test_scan(mock_writer, mock_scanner, mock_refresher, mock_id_generator):
    mock_scanner.scan_tables.return_value = [
        {'data_source': 'source.schema1.table1', 'model_name': 'source.schema1.table1'}
    ]
    mock_id_generator.generate.return_value = ['id']
    mock_scanner.scan_columns.return_value = [
        {
            'data_source': 'source.schema1.table1',
            'model_name': 'source.schema1.table1',
            'taxon_type': 'dimension',
            'data_reference': '"ID"',
            'field_map': ['id'],
            'validation_type': 'text',
            'data_type': 'CHARACTER VARYING',
        }
    ]

    scan('test-source', 'test-filter')

    assert mock_refresher.refresh_table.mock_calls == [call('schema1.table1')]
    assert mock_writer.write_scanned_model.call_count == 1


def test_scan_single_table_error(mock_writer, mock_scanner, mock_refresher, mock_id_generator):
    mock_scanner.scan_tables.return_value = [
        {'data_source': 'source.schema1.table1', 'model_name': 'source.schema1.table1'},
        {'data_source': 'source.schema1.table2', 'model_name': 'source.schema1.table2'},
    ]
    mock_id_generator.generate.return_value = ['id']
    mock_scanner.scan_columns.return_value = [
        {
            'data_source': 'source.schema1.table2',
            'model_name': 'source.schema1.table2',
            'taxon_type': 'dimension',
            'data_reference': '"ID"',
            'field_map': ['id'],
            'validation_type': 'text',
            'data_type': 'CHARACTER VARYING',
        }
    ]
    mock_refresher.refresh_table.side_effect = [Exception('test'), None]

    scan('test-source', 'test-filter')

    assert mock_writer.write_scanned_model.call_count == 1


@pytest.fixture()
def mock_physical_data_source_client():
    with patch('panoramic.cli.command.PhysicalDataSourceClient') as client_class:
        yield client_class()


def test_list_connections(mock_physical_data_source_client):
    list_connections()
    mock_physical_data_source_client.get_sources.assert_called_with('test-company')


@patch('panoramic.cli.command.echo_diff')
@patch('panoramic.cli.command.get_company_slug')
@patch('panoramic.cli.command.get_remote_state')
@patch('panoramic.cli.command.get_local_state')
@patch('panoramic.cli.command.click')
@patch('panoramic.cli.command.reconcile')
@patch.object(RemoteExecutor, '_execute')
def test_push_single_error(mock_execute, mock_reconcile, mock_click, _, __, ___, ____, capsys):
    mock_reconcile.return_value = Mock(
        actions=[Mock(description='test-description-1'), Mock(description='test-description-2'), Mock()]
    )
    mock_execute.side_effect = [
        InvalidDatasetException(Mock()),
        InvalidModelException(Mock()),
        None,
    ]

    mock_click.confirm.return_value = True  # Mock user confirming prompt

    push()

    assert mock_execute.call_count == 3
    assert capsys.readouterr().out == (
        "Loading local state...\n"
        "Fetching remote state...\n"
        "Resolving state...\n"
        "\n"
        "Error: Failed to execute action test-description-1:\n"
        "  Invalid dataset submitted\n"
        "Error: Failed to execute action test-description-2:\n"
        "  Invalid model submitted\n"
        "Updated 1/3 models, fields and datasets\n"
    )


@patch('panoramic.cli.command.get_local_state')
@patch('panoramic.cli.command.validate_orphaned_files')
@patch.object(LocalExecutor, '_execute')
def test_delete_orphaned_fields(mock_execute, mock_validate, mock_state, capsys):
    mock_state.return_value.get_objects_by_package.return_value.items.return_value = [
        ('test_dataset', ([Mock(slug='test_slug', data_source='test_dataset')], [Mock()])),
    ]

    mock_validate.return_value = [OrphanFieldFileError(field_slug='test_slug', dataset_slug='test_dataset')]

    delete_orphaned_fields(yes=True)

    assert mock_execute.call_count == 1
    assert capsys.readouterr().out == (
        "Loading local state...\n\n"
        "Fields without calculation or reference in a model in dataset test_dataset:\n"
        "  test_slug\n\n"
        "Updating local state...\n"
        "Updated 1/1 fields\n"
    )


@patch('panoramic.cli.command.get_local_state')
@patch('panoramic.cli.command.scan_fields_for_errors')
@patch.object(LocalExecutor, '_execute')
def test_scaffold_missing_files(mock_execute, mock_scan_fields_for_errors, mock_state, capsys):
    mock_state.return_value.get_objects_by_package.return_value.items.return_value = [
        (
            'test_dataset',
            (
                [],
                [Mock(data_source='db.schema.test_table', identifiers=['id'], fields=[Mock(field_map=['test_slug'])])],
            ),
        ),
    ]

    mock_scan_fields_for_errors.return_value = [sentinel.field]

    scaffold_missing_fields(yes=True)

    assert mock_execute.mock_calls == [call(Action(desired=sentinel.field))]
    assert capsys.readouterr().out == (
        "Loading local state...\n\n"
        "Fields referenced in models without definition in dataset test_dataset:\n"
        "  test_slug\n\n"
        "Scanning fields...\n"
        "Updating local state...\n"
        "Updated 1/1 fields\n"
    )



from pathlib import Path
from typing import Any, Dict

import pytest
import yaml

from panoramic.cli.errors import (
    DeprecatedAttributeWarning,
    FileMissingError,
    InvalidYamlFile,
    JsonSchemaError,
    MissingFieldFileError,
    OrphanFieldFileError,
)
from panoramic.cli.local.get import get_state
from panoramic.cli.paths import Paths, PresetFileName
from panoramic.cli.validate import (
    validate_config,
    validate_context,
    validate_local_state,
)

VALID_CONTEXT = {
    'api_version': 'v1',
    'company_slug': 'test_company',
}


INVALID_CONTEXTS = [
    # typo in version
    {
        'api_versio': 'v1',
        'company_slug': 'test_company',
    },
    # wrong type in version
    {
        'api_version': 1,
        'company_slug': 'test_company',
    },
    # typo in slug
    {'api_version': 'v1', 'company_slu': 'test_company'},
    # wrong type in slug
    {'api_version': 'v1', 'company_slug': 100},
]


@pytest.mark.parametrize('context', INVALID_CONTEXTS)
def test_validate_context_invalid(tmp_path, monkeypatch, context):
    monkeypatch.chdir(tmp_path)
    with Paths.context_file().open('w') as f:
        f.write(yaml.dump(context))

    with pytest.raises(JsonSchemaError):
        validate_context()


def test_validate_context_invalid_yaml(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    with Paths.context_file().open('w') as f:
        f.write('not:\nyaml')

    with pytest.raises(InvalidYamlFile):
        validate_context()


def test_validate_context_missing_file(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    with pytest.raises(FileMissingError):
        validate_context()


def test_validate_context_valid(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)
    with Paths.context_file().open('w') as f:
        f.write(yaml.dump(VALID_CONTEXT))

    validate_context()


VALID_CONFIG = {
    'client_id': 'test-client-id',
    'client_secret': 'test-client-secret',
}


INVALID_CONFIGS = [
    # typo in client_id
    {'client_i': 'test_client_id', 'client_secret': 'test_client_secret'},
    # wrong type in client_id
    {'client_id': 100, 'client_secret': 'test_client_secret'},
    # typo in client_secret
    {'client_id': 'test_client_id', 'client_secre': 'test_client_secret'},
    # wrong type in client_secret
    {'client_id': 'test_client_id', 'client_secret': 200},
]


@pytest.mark.parametrize('config', INVALID_CONFIGS)
def test_validate_config_invalid_env_var_set(tmp_path, monkeypatch, config):
    monkeypatch.setenv('PANO_CLIENT_ID', 'test-client_id')
    monkeypatch.setenv('PANO_CLIENT_SECRET', 'test-client-secret')
    monkeypatch.setattr(Path, 'home', lambda: tmp_path)

    Paths.config_dir().mkdir()
    with Paths.config_file().open('w') as f:
        f.write(yaml.dump(config))

    validate_config()


@pytest.mark.parametrize('config', INVALID_CONFIGS)
def test_validate_config_invalid(tmp_path, monkeypatch, config):
    monkeypatch.delenv('PANO_CLIENT_ID', raising=False)
    monkeypatch.delenv('PANO_CLIENT_SECRET', raising=False)
    monkeypatch.setattr(Path, 'home', lambda: tmp_path)

    Paths.config_dir().mkdir()
    with Paths.config_file().open('w') as f:
        f.write(yaml.dump(config))

    with pytest.raises(JsonSchemaError):
        validate_config()


def test_validate_config_invalid_yaml(tmp_path, monkeypatch):
    monkeypatch.delenv('PANO_CLIENT_ID', raising=False)
    monkeypatch.delenv('PANO_CLIENT_SECRET', raising=False)
    monkeypatch.setattr(Path, 'home', lambda: tmp_path)

    Paths.config_dir().mkdir()
    with Paths.config_file().open('w') as f:
        f.write('not:\nyaml')

    with pytest.raises(InvalidYamlFile):
        validate_config()


def test_validate_config_missing_file(tmp_path, monkeypatch):
    monkeypatch.delenv('PANO_CLIENT_ID', raising=False)
    monkeypatch.delenv('PANO_CLIENT_SECRET', raising=False)
    monkeypatch.setattr(Path, 'home', lambda: tmp_path)

    with pytest.raises(FileMissingError):
        validate_config()


def test_validate_config_valid(tmp_path, monkeypatch):
    monkeypatch.setattr(Path, 'home', lambda: tmp_path)

    Paths.config_dir().mkdir()
    with Paths.config_file().open('w') as f:
        f.write(yaml.dump(VALID_CONFIG))

    validate_config()


VALID_DATASET = {
    'dataset_slug': 'test-slug',
    'display_name': 'Test Name',
    'api_version': 'v1',
}
VALID_MODEL_MINIMAL: Dict[str, Any] = {
    'model_name': 'sf.db.schema.table1',
    'data_source': 'sf.db.schema.table1',
    'api_version': 'v1',
}
VALID_MODEL_FULL = {
    'model_name': 'sf.db.schema.table1',
    'data_source': 'sf.db.schema.table1',
    'api_version': 'v1',
    'identifiers': ['ad_id'],
    'joins': [
        {'to_model': 'sf.db.schema.table2', 'join_type': 'left', 'relationship': 'many_to_one', 'fields': ['ad_id']},
        {'to_model': 'sf.db.schema.table3', 'join_type': 'right', 'relationship': 'one_to_one', 'fields': ['ad_id']},
        {'to_model': 'sf.db.schema.table4', 'join_type': 'inner', 'relationship': 'one_to_one', 'fields': ['ad_id']},
        {'to_model': 'sf.db.schema.table5', 'join_type': 'inner', 'relationship': 'many_to_one', 'fields': ['ad_id']},
    ],
    'fields': [
        # has field_map
        {
            'data_reference': '"ad_id"',
            'field_map': ['ad_id'],
        },
    ],
}
VALID_FIELD_MINIMAL: Dict[str, Any] = {
    'api_version': 'v1',
    'slug': 'field_slug',
    'group': 'group',
    'display_name': 'Display name',
    'data_type': 'data_type',
    'field_type': 'field_type',
}

VALID_FIELD_FULL: Dict[str, Any] = {
    'slug': 'full_field_slug',
    'group': 'group',
    'display_name': 'Full display name',
    'data_type': 'data_type',
    'field_type': 'field_type',
    'calculation': 'calculation',
    'aggregation': {'type': 'sum'},
    'display_format': 'display_format',
    'description': 'description',
}


INVALID_DATASETS = [
    # typo in dataset_slug
    {
        'dataset_slu': 'test-slug',
        'display_name': 'Test Name',
        'api_version': 'v1',
    },
    # wrong type in dataset_slug
    {
        'dataset_slug': 100,
        'display_name': 'Test Name',
        'api_version': 'v1',
    },
    # typo in display_name
    {
        'dataset_slug': 'test-slug',
        'display_nam': 'Test Name',
        'api_version': 'v1',
    },
    # wrong type in display_name
    {
        'dataset_slug': 'test-slug',
        'display_name': 200,
        'api_version': 'v1',
    },
    # typo in api_version
    {
        'dataset_slug': 'test-slug',
        'display_name': 'Test Name',
        'api_versio': 'v1',
    },
    # wrong type in api_version
    {
        'dataset_slug': 'test-slug',
        'display_name': 'Test Name',
        'api_version': 1,
    },
]


INVALID_MODELS = [
    # typo in model_name
    {**VALID_MODEL_MINIMAL, 'model_nae': 'sf.db.schema.table1'},
    # wrong type in model_name
    {**VALID_MODEL_MINIMAL, 'model_name': 100},
    # typo in data_source
    {**VALID_MODEL_MINIMAL, 'data_sourc': 'sf.db.schema.table1'},
    # wrong type in data_source
    {**VALID_MODEL_MINIMAL, 'data_source': 200},
    # typo in api_version
    {
        **VALID_MODEL_MINIMAL,
        'api_versio': 'v1',
    },
    # wrong type in api_version
    {
        **VALID_MODEL_MINIMAL,
        'api_version': 1,
    },
    # typo in identifiers
    {
        **VALID_MODEL_MINIMAL,
        'identifirs': ['ad_id'],
    },
    # wrong type in identifiers
    {
        **VALID_MODEL_MINIMAL,
        'identifiers': [100],
    },
    # typo in joins
    {
        **VALID_MODEL_MINIMAL,
        'jons': [
            {
                'to_model': 'sf.db.schema.table2',
                'join_type': 'left',
                'relationship': 'many_to_one',
                'fields': ['ad_id'],
            },
        ],
    },
    # invalid join type
    {
        **VALID_MODEL_MINIMAL,
        'joins': [
            {
                'to_model': 'sf.db.schema.table2',
                'join_type': 'middle',
                'relationship': 'many_to_one',
                'fields': ['ad_id'],
            },
        ],
    },
    # invalid relationship
    {
        **VALID_MODEL_MINIMAL,
        'joins': [
            {
                'to_model': 'sf.db.schema.table2',
                'join_type': 'left',
                'relationship': 'single_to_single',
                'fields': ['ad_id'],
            },
        ],
    },
    # no "to model" set
    {
        **VALID_MODEL_MINIMAL,
        'joins': [
            {
                'join_type': 'left',
                'relationship': 'many_to_one',
                'fields': ['ad_id'],
            },
        ],
    },
    # no join fields set
    {
        **VALID_MODEL_MINIMAL,
        'joins': [
            {
                'to_model': 'sf.db.schema.table2',
                'join_type': 'left',
                'relationship': 'many_to_one',
                'fields': [],
            },
        ],
    },
    # wrong type in joins
    {
        **VALID_MODEL_MINIMAL,
        'joins': [100],
    },
    # typo in fields
    {
        **VALID_MODEL_MINIMAL,
        'felds': [
            {
                'data_reference': '"campaign_id"',
                'field_map': ['campaign_id'],
                'data_type': 'CHARACTER VARYING',
            },
        ],
    },
    # wrong type in fields
    {
        **VALID_MODEL_MINIMAL,
        'fields': [100],
    },
    # field_map not set
    {
        **VALID_MODEL_MINIMAL,
        'fields': [
            {
                'data_reference': '"campaign_id"',
                'data_type': 'CHARACTER VARYING',
            },
        ],
    },
    # data_type not set
    {
        **VALID_MODEL_MINIMAL,
        'fields': [
            {
                'data_reference': '"campaign_id"',
            },
        ],
    },
    # data_reference not set
    {
        **VALID_MODEL_MINIMAL,
        'fields': [
            {
                'data_type': 'CHARACTER VARYING',
            },
        ],
    },
]

INVALID_FIELDS = [
    # typo in slug
    {**VALID_FIELD_MINIMAL, 'sulg': 'should_be_slug'},
    # slug not set
    {k: v for k, v in VALID_FIELD_MINIMAL.items() if k != 'slug'},
    # wrong type in slug
    {**VALID_FIELD_MINIMAL, 'slug': 123},
    # group not set
    {k: v for k, v in VALID_FIELD_MINIMAL.items() if k != 'group'},
    # field_type not set
    {k: v for k, v in VALID_FIELD_MINIMAL.items() if k != 'field_type'},
    # data_type not set
    {k: v for k, v in VALID_FIELD_MINIMAL.items() if k != 'data_type'},
    # typo in api_version
    {**VALID_FIELD_MINIMAL, 'api_versio': 'v1'},
    # wrong type in api_version
    {**VALID_FIELD_MINIMAL, 'api_version': 1},
    # wrong type in calculation
    {**VALID_FIELD_MINIMAL, 'calculation': 1},
    # wrong type in aggregation
    {**VALID_FIELD_MINIMAL, 'aggregation': 1},
    {**VALID_FIELD_MINIMAL, 'aggregation': ''},
    # Invalid aggregation object
    {**VALID_FIELD_MINIMAL, 'aggregation': {'nope': 1}},
    {**VALID_FIELD_MINIMAL, 'aggregation': {'type': 1}},
    {**VALID_FIELD_MINIMAL, 'aggregation': {'type': {}}},
    {**VALID_FIELD_MINIMAL, 'aggregation': {'type': 'sum', 'params': '1232'}},
    {**VALID_FIELD_MINIMAL, 'aggregation': {'type': 'sum', 'params': {}, 'nope': 1}},
    # wrong type in display_format
    {**VALID_FIELD_MINIMAL, 'display_format': 1},
]


def test_validate_local_state_valid(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    global_fields_dir = Paths.fields_dir(tmp_path)
    global_fields_dir.mkdir()

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()
    dataset_fields_dir = Paths.fields_dir(dataset_dir)
    dataset_fields_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    model1 = {**VALID_MODEL_MINIMAL, 'model_name': 'sf.db.schema.table1'}
    model2 = {
        **VALID_MODEL_MINIMAL,
        'model_name': 'sf.db.schema.table2',
        'fields': [{'field_map': ['field_slug'], 'data_reference': '"FIELD_SLUG"'}],
    }

    with (dataset_dir / 'test_model-1.model.yaml').open('w') as f:
        f.write(yaml.dump(model1))

    with (dataset_dir / 'test_model-2.model.yaml').open('w') as f:
        f.write(yaml.dump(model2))

    with (global_fields_dir / 'company_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_FULL))

    with (dataset_fields_dir / 'first_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_MINIMAL))

    errors = validate_local_state()
    assert len(errors) == 0

    state = get_state()
    assert len(state.models) == 2
    assert len(state.data_sources) == 1
    assert len(state.fields) == 2


@pytest.mark.parametrize('dataset', INVALID_DATASETS)
def test_validate_local_state_invalid_dataset(tmp_path, monkeypatch, dataset):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(dataset))

    errors = validate_local_state()
    assert len(errors) == 1


@pytest.mark.parametrize('invalid_field', INVALID_FIELDS)
def test_validate_local_state_invalid_dataset_scoped_field(tmp_path, monkeypatch, invalid_field):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    field_dir = Paths.fields_dir(dataset_dir)
    field_dir.mkdir()

    with (field_dir / 'first_field.field.yaml').open('w') as f:
        f.write(yaml.dump(invalid_field))

    errors = validate_local_state()
    assert len(errors) == 1


@pytest.mark.parametrize('invalid_field', INVALID_FIELDS)
def test_validate_local_state_duplicate_dataset_scoped_field(tmp_path, monkeypatch, invalid_field):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    with (dataset_dir / 'test_model.model.yaml').open('w') as f:
        f.write(
            yaml.dump(
                {**VALID_MODEL_MINIMAL, 'fields': [{'field_map': ['field_slug'], 'data_reference': '"FIELD_SLUG"'}]}
            )
        )

    field_dir = Paths.fields_dir(dataset_dir)
    field_dir.mkdir()

    with (field_dir / 'first_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_MINIMAL))

    with (field_dir / 'duplicate.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_MINIMAL))

    errors = validate_local_state()
    assert len(errors) == 1


@pytest.mark.parametrize('invalid_field', INVALID_FIELDS)
def test_validate_local_state_invalid_company_scoped_field(tmp_path, monkeypatch, invalid_field):
    monkeypatch.chdir(tmp_path)

    global_field_dir = tmp_path / 'fields'
    global_field_dir.mkdir()

    with (global_field_dir / 'a_field.field.yaml').open('w') as f:
        f.write(yaml.dump(invalid_field))

    errors = validate_local_state()
    assert len(errors) == 1


@pytest.mark.parametrize('invalid_field', INVALID_FIELDS)
def test_validate_local_state_duplicate_company_scoped_field(tmp_path, monkeypatch, invalid_field):
    monkeypatch.chdir(tmp_path)

    global_field_dir = tmp_path / 'fields'
    global_field_dir.mkdir()

    with (global_field_dir / 'a_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_FULL))

    with (global_field_dir / 'duplicate_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_FULL))

    errors = validate_local_state()
    assert len(errors) == 1


@pytest.mark.parametrize('model', INVALID_MODELS)
def test_validate_local_state_invalid_models(tmp_path, monkeypatch, model):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    with (dataset_dir / 'test_model.model.yaml').open('w') as f:
        f.write(yaml.dump(model))

    errors = validate_local_state()
    assert len(errors) == 1


def test_validate_local_state_duplicate_model_names(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    with (dataset_dir / 'test_model-1.model.yaml').open('w') as f:
        f.write(yaml.dump(VALID_MODEL_MINIMAL))

    with (dataset_dir / 'test_model-2.model.yaml').open('w') as f:
        f.write(yaml.dump(VALID_MODEL_MINIMAL))

    errors = validate_local_state()
    assert len(errors) == 1


def test_validate_local_state_missing_field_file(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    with (dataset_dir / 'model1.model.yaml').open('w') as f:
        f.write(
            yaml.dump(
                {
                    **VALID_MODEL_MINIMAL,
                    'fields': [{'data_reference': '"COLUMN1"', 'field_map': ['field_slug', 'field_slug_2']}],
                }
            )
        )

    field_dir = Paths.fields_dir(dataset_dir)
    field_dir.mkdir()

    with (field_dir / 'field_slug.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_MINIMAL))

    errors = validate_local_state()
    assert errors == [
        MissingFieldFileError(
            field_slug='field_slug_2',
            dataset_slug='test_dataset',
            data_source='sf.db.schema.table1',
            data_reference='"COLUMN1"',
            identifier=False,
        )
    ]


def test_validate_local_state_deprecated_attribute(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    with (dataset_dir / 'test_model.model.yaml').open('w') as f:
        f.write(
            yaml.dump(
                {
                    **VALID_MODEL_MINIMAL,
                    'fields': [
                        {
                            'data_type': 'CHARACTER VARYING',
                            'field_map': ['field_slug'],
                            'data_reference': '"FIELD_SLUG"',
                        }
                    ],
                }
            )
        )

    Paths.fields_dir(dataset_dir).mkdir()
    with (Paths.fields_dir(dataset_dir) / 'test_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_MINIMAL))

    errors = validate_local_state()
    assert errors == [DeprecatedAttributeWarning(attribute='data_type', path=dataset_dir / 'test_model.model.yaml')]


def test_validate_local_state_orphan_field_files(tmp_path, monkeypatch):
    monkeypatch.chdir(tmp_path)

    dataset_dir = tmp_path / 'test_dataset'
    dataset_dir.mkdir()

    with (dataset_dir / PresetFileName.DATASET_YAML.value).open('w') as f:
        f.write(yaml.dump(VALID_DATASET))

    with (dataset_dir / 'test_model.model.yaml').open('w') as f:
        f.write(
            yaml.dump(
                {
                    **VALID_MODEL_FULL,
                    'fields': [{'field_map': ['field_slug'], 'data_reference': '"FIELD_SLUG"'}],
                }
            )
        )

    Paths.fields_dir(dataset_dir).mkdir()
    with (Paths.fields_dir(dataset_dir) / 'test_field.field.yaml').open('w') as f:
        f.write(yaml.dump(VALID_FIELD_MINIMAL))

    with (Paths.fields_dir(dataset_dir) / 'calculated_field.field.yaml').open('w') as f:
        f.write(yaml.dump({**VALID_FIELD_MINIMAL, 'slug': 'calculated_slug', 'calculation': '2+2'}))

    with (Paths.fields_dir(dataset_dir) / 'orphan_field.field.yaml').open('w') as f:
        f.write(yaml.dump({**VALID_FIELD_MINIMAL, 'slug': 'orphan_slug'}))

    errors = validate_local_state()

    assert errors == [OrphanFieldFileError(field_slug='orphan_slug', dataset_slug='test_dataset')]



import functools

from panoramic.cli.file_utils import read_yaml
from panoramic.cli.paths import Paths


@functools.lru_cache()
def get_company_slug() -> str:
    """Return company slug from context."""
    return read_yaml(Paths.context_file())['company_slug']



from typing import Any, Dict, List, Optional

from panoramic.auth.client import OAuth2Client
from panoramic.cli.clients import VersionedClient
from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.config.source import get_base_url


class PhysicalDataSourceClient(OAuth2Client, VersionedClient):

    """Physical data source HTTP API client."""

    base_url: str

    def __init__(
        self, base_url: Optional[str] = None, client_id: Optional[str] = None, client_secret: Optional[str] = None
    ):
        client_id = client_id if client_id is not None else get_client_id()
        client_secret = client_secret if client_secret is not None else get_client_secret()

        self.base_url = base_url if base_url is not None else get_base_url()

        super().__init__(client_id, client_secret)

    def get_sources(self, company_slug: str) -> List[Dict[str, Any]]:
        response = self.session.get(self.base_url, params={'company_slug': company_slug}, timeout=30)
        response.raise_for_status()
        return response.json()['data']



import json
import logging
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

from panoramic.auth import OAuth2Client
from panoramic.cli.clients import VersionedClient
from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.config.field import get_base_url

logger = logging.getLogger(__name__)


class Aggregation:

    type: str
    params: Optional[Dict[str, Any]]

    def __init__(
        self,
        *,
        type: str,
        params: Optional[Dict[str, Any]],
    ):
        self.type = type
        self.params = params

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'Aggregation':
        return cls(type=inputs['type'], params=inputs.get('params'))

    def to_dict(self) -> Dict[str, Any]:
        data: Dict[str, Any] = {'type': self.type}
        if self.params is not None:
            data['params'] = self.params
        return data

    def __hash__(self) -> int:
        return hash(json.dumps(self.to_dict()))

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Aggregation):
            return False

        return self.to_dict() == o.to_dict()


class Field:

    slug: str
    group: str
    display_name: str
    data_type: str
    field_type: str
    description: Optional[str]
    data_source: Optional[str]
    calculation: Optional[str]
    aggregation: Optional[Aggregation]
    display_format: Optional[str]

    def __init__(
        self,
        *,
        slug: str,
        group: str,
        display_name: str,
        data_type: str,
        field_type: str,
        description: Optional[str],
        data_source: Optional[str],
        calculation: Optional[str],
        aggregation: Optional[Aggregation],
        display_format: Optional[str],
    ):
        self.slug = slug
        self.group = group
        self.display_name = display_name
        self.data_type = data_type
        self.field_type = field_type
        self.description = description
        self.data_source = data_source
        self.calculation = calculation
        self.aggregation = aggregation
        self.display_format = display_format

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'Field':
        aggregation = inputs.get('aggregation')
        return cls(
            slug=inputs['slug'],
            group=inputs['group'],
            display_name=inputs['display_name'],
            data_type=inputs['data_type'],
            field_type=inputs['field_type'],
            description=inputs.get('description'),
            data_source=inputs.get('data_source'),
            aggregation=Aggregation.from_dict(aggregation) if aggregation is not None else None,
            calculation=inputs.get('calculation'),
            display_format=inputs.get('display_format'),
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            'slug': self.slug,
            'group': self.group,
            'display_name': self.display_name,
            'data_type': self.data_type,
            'field_type': self.field_type,
            'description': self.description,
            'calculation': self.calculation,
            'aggregation': self.aggregation.to_dict() if self.aggregation is not None else None,
            'data_source': self.data_source,
            'display_format': self.display_format,
        }

    def __hash__(self) -> int:
        return hash(
            (
                self.slug,
                self.group,
                self.display_name,
                self.data_type,
                self.field_type,
                self.description,
                self.data_source,
                self.calculation,
                self.aggregation,
                self.display_format,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Field):
            return False

        return self.to_dict() == o.to_dict()


class FieldClient(OAuth2Client, VersionedClient):

    """Model field HTTP API client."""

    base_url: str
    _base_url_with_trailing_slash: str

    def __init__(
        self,
        base_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        client_id = client_id if client_id is not None else get_client_id()
        client_secret = client_secret if client_secret is not None else get_client_secret()

        # Since we need to request api/virtual?company_id=x and api/virtual/slug?company_id=1
        # the base gets corrected to not include trailing slash
        #
        # Check https://stackoverflow.com/questions/10893374/python-confusions-with-urljoin for more context
        self.base_url = base_url if base_url is not None else get_base_url()
        if self.base_url[-1] == '/':
            self._base_url_with_trailing_slash = self.base_url
            self.base_url = self.base_url[0:-1]
        else:
            # base_url is in it's correct form - without trailing slash
            self._base_url_with_trailing_slash = self.base_url + '/'

        super().__init__(client_id, client_secret)

    def upsert_fields(self, company_slug: str, fields: List[Field], data_source: Optional[str] = None):
        params = {'company_slug': company_slug, 'virtual_data_source': data_source}
        response = self.session.post(
            self.base_url, params=params, json=[field.to_dict() for field in fields], timeout=30
        )
        response.raise_for_status()

    def get_fields(
        self, company_slug: str, data_source: Optional[str] = None, offset: int = 0, limit: int = 100
    ) -> List[Field]:
        """Retrieve all fields in a given source and optionally a datasource."""
        logger.debug(f'Listing fields for company: {company_slug} source: {data_source}')
        params = {'company_slug': company_slug, 'virtual_data_source': data_source, 'offset': offset, 'limit': limit}
        response = self.session.get(self.base_url, params=params)
        response.raise_for_status()
        response_json = response.json()['data']
        return [Field.from_dict(d) for d in response_json]

    def delete_fields(self, company_slug: str, slugs: List[str]):
        """Delete a field with a given name."""
        url = urljoin(self._base_url_with_trailing_slash, 'delete')
        logger.debug(f'Deleting fields : {", ".join(slugs)}')
        response = self.session.post(url, json=slugs, params={'company_slug': company_slug}, timeout=30)
        response.raise_for_status()



from unittest.mock import Mock, call, patch, sentinel

import pytest
from requests.exceptions import RequestException

from panoramic.cli.errors import (
    DatasetWriteException,
    FieldWriteException,
    InvalidDatasetException,
    InvalidFieldException,
    InvalidModelException,
    ModelWriteException,
)
from panoramic.cli.field_mapper import map_field_slug_from_local
from panoramic.cli.remote.writer import RemoteWriter


@pytest.fixture
def mock_model_client():
    with patch('panoramic.cli.remote.writer.ModelClient', autospec=True) as client:
        yield client()


@pytest.fixture
def mock_vds_client():
    with patch('panoramic.cli.remote.writer.VirtualDataSourceClient', autospec=True) as client:
        yield client()


@pytest.fixture
def mock_field_client():
    with patch('panoramic.cli.remote.writer.FieldClient', autospec=True) as client:
        yield client()


@patch('panoramic.cli.remote.writer.map_data_source_from_local')
def test_writer_write_data_source_invalid_exception(_, mock_vds_client):
    mock_vds_client.upsert_virtual_data_source.side_effect = RequestException('test', response=Mock(status_code=400))

    with pytest.raises(InvalidDatasetException):
        RemoteWriter('test_company').write_data_source(Mock())


@patch('panoramic.cli.remote.writer.map_data_source_from_local')
def test_writer_write_data_source_exception(_, mock_vds_client):
    mock_vds_client.upsert_virtual_data_source.side_effect = RequestException('test')

    with pytest.raises(DatasetWriteException):
        RemoteWriter('test_company').write_data_source(Mock())


@patch('panoramic.cli.remote.writer.map_data_source_from_local')
def test_writer_write_data_source(mock_map, mock_vds_client):
    mock_map.return_value = sentinel.remote_dataset

    RemoteWriter('test_company').write_data_source(Mock())

    assert mock_vds_client.upsert_virtual_data_source.mock_calls == [call('test_company', sentinel.remote_dataset)]


@patch('panoramic.cli.remote.writer.map_model_from_local')
def test_writer_write_model_invalid_exception(_, mock_model_client):
    mock_model_client.upsert_model.side_effect = RequestException('test', response=Mock(status_code=400))

    with pytest.raises(InvalidModelException):
        RemoteWriter('test_company').write_model(Mock())


@patch('panoramic.cli.remote.writer.map_model_from_local')
def test_writer_write_model_exception(_, mock_model_client):
    mock_model_client.upsert_model.side_effect = RequestException('test')

    with pytest.raises(ModelWriteException):
        RemoteWriter('test_company').write_model(Mock())


@patch('panoramic.cli.remote.writer.map_model_from_local')
def test_writer_model(mock_map, mock_model_client):
    mock_map.return_value = sentinel.remote_model

    RemoteWriter('test_company').write_model(Mock(virtual_data_source='test_dataset'))

    assert mock_model_client.upsert_model.mock_calls == [call('test_dataset', 'test_company', sentinel.remote_model)]


def test_writer_delete_model_exception(mock_model_client):
    model = Mock(model_name='test_model', virtual_data_source='test_dataset')
    mock_model_client.delete_model.side_effect = RequestException('test')

    with pytest.raises(ModelWriteException):
        RemoteWriter('test_company').delete_model(model)

    assert mock_model_client.delete_model.mock_calls == [call('test_dataset', 'test_company', 'test_model')]


def test_writer_delete_data_source_exception(mock_vds_client):
    data_source = Mock(dataset_slug='test_dataset')
    mock_vds_client.delete_virtual_data_source.side_effect = RequestException('test')

    with pytest.raises(DatasetWriteException):
        RemoteWriter('test_company').delete_data_source(data_source)


def test_writer_delete_field_exception(mock_field_client):
    field = Mock(slug='slug', data_source='data_source')

    mock_field_client.delete_fields.side_effect = RequestException('test')

    with pytest.raises(FieldWriteException):
        RemoteWriter('test_company').delete_field(field)


def test_writer_delete_model(mock_model_client):
    model = Mock(model_name='test_model', virtual_data_source='test_dataset')

    RemoteWriter('test_company').delete_model(model)

    assert mock_model_client.delete_model.mock_calls == [call('test_dataset', 'test_company', 'test_model')]


def test_writer_delete_data_source(mock_vds_client):
    data_source = Mock(dataset_slug='test_dataset')

    RemoteWriter('test_company').delete_data_source(data_source)

    assert mock_vds_client.delete_virtual_data_source.mock_calls == [call('test_company', 'test_dataset')]


def test_writer_delete_field(mock_field_client):
    field = Mock(slug='slug', data_source='data_source')
    remote_slug = map_field_slug_from_local('slug', 'data_source')

    RemoteWriter('test_company').delete_field(field)

    mock_field_client.delete_fields.assert_called_with(company_slug='test_company', slugs=[remote_slug])


@patch('panoramic.cli.remote.writer.map_field_from_local')
def test_writer_write_field(mock__map_field_from_local, mock_field_client):
    mock__map_field_from_local.return_value = sentinel.remote_field

    RemoteWriter('test_company').write_field(Mock())

    assert mock_field_client.upsert_fields.mock_calls == [
        call(company_slug='test_company', fields=[sentinel.remote_field])
    ]


@patch('panoramic.cli.remote.writer.map_field_from_local')
def test_writer_write_field_invalid_exception(_, mock_field_client):
    mock_field_client.upsert_fields.side_effect = RequestException('test', response=Mock(status_code=400))

    with pytest.raises(InvalidFieldException):
        RemoteWriter('test_company').write_field(Mock())


@patch('panoramic.cli.remote.writer.map_field_from_local')
def test_writer_write_field_source_exception(_, mock_field_client):
    mock_field_client.upsert_fields.side_effect = RequestException('test')

    with pytest.raises(FieldWriteException):
        RemoteWriter('test_company').write_field(Mock())



import itertools
from typing import Iterable, Sequence

from panoramic.cli.pano_model import Actionable
from panoramic.cli.state import Action, ActionList, VirtualState


def _reconcile_actions(
    current_entities: Sequence[Actionable], desired_entities: Sequence[Actionable]
) -> Iterable[Action]:
    """Create actions that get us from current to desired state."""
    current_by_id = {s.id: s for s in current_entities}
    desired_by_id = {s.id: s for s in desired_entities}

    # delete what is not desired but exists
    ids_to_delete = current_by_id.keys() - desired_by_id.keys()
    # create what is desired but doesn't exist
    ids_to_create = desired_by_id.keys() - current_by_id.keys()
    # update what is desired and exists
    ids_to_update = current_by_id.keys() & desired_by_id.keys()

    for item_id in ids_to_delete:
        yield Action(current=current_by_id[item_id])

    for item_id in ids_to_create:
        yield Action(desired=desired_by_id[item_id])

    for item_id in ids_to_update:
        if current_by_id[item_id] != desired_by_id[item_id]:
            yield Action(current=current_by_id[item_id], desired=desired_by_id[item_id])


def reconcile(current_state: VirtualState, desired_state: VirtualState) -> ActionList:
    """Create actions that get us from current state to desired state."""
    actions = list(
        itertools.chain(
            _reconcile_actions(current_state.data_sources, desired_state.data_sources),
            _reconcile_actions(current_state.models, desired_state.models),
            _reconcile_actions(current_state.fields, desired_state.fields),
        )
    )
    return ActionList(actions=actions)



from unittest.mock import Mock, patch

import pytest
from requests.models import HTTPError

from panoramic.cli.errors import RefreshException, SourceNotFoundException
from panoramic.cli.metadata.client import TERMINAL_STATES, JobState
from panoramic.cli.refresh import Refresher


@patch('panoramic.cli.refresh.MetadataClient')
def test_refresher_completed(mock_client):
    mock_client.return_value.create_refresh_job.return_value = 'test-job-id'
    mock_client.return_value.wait_for_terminal_state.return_value = JobState.COMPLETED

    Refresher('test-company', 'test-source').refresh_table('table-name')


@pytest.mark.parametrize('final_state', list(TERMINAL_STATES - {JobState.COMPLETED}))
@patch('panoramic.cli.refresh.MetadataClient')
def test_refresher_non_completed(mock_client, final_state):
    mock_client.return_value.create_refresh_job.return_value = 'test-job-id'
    mock_client.return_value.wait_for_terminal_state.return_value = final_state

    with pytest.raises(RefreshException):
        Refresher('test-company', 'test-source').refresh_table('table-name')


@patch('panoramic.cli.refresh.MetadataClient')
def test_refresher_not_found(mock_client):
    mock_client.return_value.create_refresh_job.side_effect = HTTPError(response=Mock(status_code=404))

    with pytest.raises(SourceNotFoundException):
        Refresher('test-company', 'test-source').refresh_table('table-name')


@patch('panoramic.cli.refresh.MetadataClient')
def test_refresher_generic_error(mock_client):
    mock_client.return_value.create_refresh_job.side_effect = HTTPError(response=Mock(status_code=500))

    with pytest.raises(RefreshException):
        Refresher('test-company', 'test-source').refresh_table('table-name')



import distutils.util
import os
from typing import Optional

import requests

from panoramic.cli.supported_version import fetch_analytics_write_key


def is_enabled() -> bool:
    try:
        enabled = os.environ.get('PANO_ANALYTICS_ENABLED', 'true')
        return distutils.util.strtobool(enabled)
    except ValueError:
        return True


def get_write_key() -> Optional[str]:
    """Get Segment integration source write key."""
    write_key = os.environ.get("PANO_ANALYTICS_WRITE_KEY", None)
    if write_key is not None:
        return write_key

    try:
        write_key = fetch_analytics_write_key()
        if write_key is not None:
            return write_key
    except requests.exceptions.RequestException:
        pass

    return None



from abc import ABC
from typing import Any, Dict, List, Optional, Tuple, Union

from panoramic.cli.field.client import Aggregation


class Actionable(ABC):

    """Interface for object that you can perform actions on."""

    package: Optional[str] = None

    file_name: Optional[str] = None

    @property
    def id(_) -> Any:
        raise NotImplementedError('id not implemented for base class')

    def to_dict(self) -> Dict[str, Any]:
        raise NotImplementedError('to_dict not implemented for base class')

    def __hash__(self) -> int:
        raise NotImplementedError('__hash__ not implemented for base class')

    def __eq__(self, o: object) -> bool:
        raise NotImplementedError('__eq__ not implemented for base class')


class PanoModelField:
    """Field stored on a model."""

    field_map: List[str]
    data_reference: str

    def __init__(self, *, field_map: List[str], data_reference: str):
        self.field_map = field_map
        self.data_reference = data_reference

    def to_dict(self) -> Dict[str, Any]:
        return {
            'field_map': self.field_map,
            'data_reference': self.data_reference,
        }

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'PanoModelField':
        return cls(
            field_map=inputs['field_map'],
            data_reference=inputs['data_reference'],
        )

    def identifier(self) -> str:
        return f'{self.data_reference}_{",".join(sorted(self.field_map))}'

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.field_map),
                self.data_reference,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PanoModelField):
            return False

        return self.to_dict() == o.to_dict()


class PanoField(Actionable):
    """Field definition."""

    API_VERSION = 'v1'

    slug: str
    group: str
    display_name: str
    data_type: str
    field_type: str
    description: Optional[str]
    data_source: Optional[str]
    calculation: Optional[str]
    display_format: Optional[str]
    aggregation: Optional[Aggregation]

    def __init__(
        self,
        *,
        slug: str,
        group: str,
        display_name: str,
        data_type: str,
        field_type: str,
        description: Optional[str] = None,
        data_source: Optional[str] = None,
        calculation: Optional[str] = None,
        display_format: Optional[str] = None,
        # TODO: create Pano class for aggregation
        aggregation: Optional[Aggregation] = None,
        package: Optional[str] = None,
        file_name: Optional[str] = None,
    ):
        self.slug = slug
        self.group = group
        self.display_name = display_name
        self.data_type = data_type
        self.field_type = field_type
        self.description = description
        self.data_source = data_source
        self.calculation = calculation
        self.display_format = display_format
        self.aggregation = aggregation
        self.package = package
        self.file_name = file_name

    def to_dict(self) -> Dict[str, Any]:
        # data_source is not persisted to YAML
        return {k: v for k, v in self._to_internal_dict().items() if k != 'data_source'}

    def _to_internal_dict(self) -> Dict[str, Any]:
        data: Dict[str, Any] = {
            'api_version': self.API_VERSION,
            'slug': self.slug,
            'group': self.group,
            'display_name': self.display_name,
            'data_type': self.data_type,
            'field_type': self.field_type,
            'data_source': self.data_source,
        }

        if self.description is not None:
            data['description'] = self.description
        if self.calculation is not None:
            data['calculation'] = self.calculation
        if self.aggregation is not None:
            data['aggregation'] = self.aggregation.to_dict()
        if self.display_format is not None:
            data['display_format'] = self.display_format

        return data

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'PanoField':
        aggregation = inputs.get('aggregation')
        return cls(
            slug=inputs['slug'],
            group=inputs['group'],
            display_name=inputs['display_name'],
            data_type=inputs['data_type'],
            field_type=inputs['field_type'],
            description=inputs.get('description'),
            data_source=inputs.get('data_source'),
            aggregation=Aggregation.from_dict(aggregation) if aggregation is not None else None,
            calculation=inputs.get('calculation'),
            display_format=inputs.get('display_format'),
            package=inputs.get('package'),
            file_name=inputs.get('file_name'),
        )

    @property
    def id(self) -> Union[Tuple[str], Tuple[str, str]]:
        if self.data_source is not None:
            return (self.data_source, self.slug)
        else:
            return (self.slug,)

    def __hash__(self) -> int:
        return hash(
            (
                self.slug,
                self.group,
                self.display_name,
                self.data_type,
                self.field_type,
                self.description,
                self.data_source,
                self.calculation,
                self.display_format,
                self.aggregation,
                self.package,
                self.file_name,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PanoField):
            return False

        return self._to_internal_dict() == o._to_internal_dict()


class PanoModelJoin:
    """Represent joins on other models."""

    fields: List[str]
    join_type: str
    relationship: str
    to_model: str

    def __init__(self, *, fields: List[str], join_type: str, relationship: str, to_model: str):
        self.fields = fields
        self.join_type = join_type
        self.relationship = relationship
        self.to_model = to_model

    def to_dict(self) -> Dict[str, Any]:
        return {
            'fields': self.fields,
            'join_type': self.join_type,
            'relationship': self.relationship,
            'to_model': self.to_model,
        }

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'PanoModelJoin':
        return cls(
            fields=inputs['fields'],
            join_type=inputs['join_type'],
            relationship=inputs['relationship'],
            to_model=inputs['to_model'],
        )

    def identifier(self) -> str:
        return f'{self.join_type}_{self.relationship}_{self.to_model}_{",".join(sorted(self.fields))}'

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.fields),
                self.join_type,
                self.relationship,
                self.to_model,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PanoModelJoin):
            return False

        return self.to_dict() == o.to_dict()


class PanoModel(Actionable):
    """Model representing some table."""

    API_VERSION = 'v1'

    model_name: str
    data_source: str
    fields: List[PanoModelField]
    joins: List[PanoModelJoin]
    identifiers: List[str]
    virtual_data_source: Optional[str]

    def __init__(
        self,
        *,
        model_name: str,
        data_source: str,
        fields: List[PanoModelField],
        joins: List[PanoModelJoin],
        identifiers: List[str],
        virtual_data_source: Optional[str] = None,
        package: Optional[str] = None,
        file_name: Optional[str] = None,
    ):
        self.model_name = model_name
        self.data_source = data_source
        self.fields = fields
        self.joins = joins
        self.identifiers = identifiers
        self.virtual_data_source = virtual_data_source
        self.package = package
        self.file_name = file_name

    @property
    def id(self) -> Tuple[Optional[str], str]:
        return (self.virtual_data_source, self.model_name)

    def to_dict(self) -> Dict[str, Any]:
        return {
            'api_version': self.API_VERSION,
            'model_name': self.model_name,
            'data_source': self.data_source,
            'fields': [x.to_dict() for x in sorted(self.fields, key=lambda field: field.identifier())],
            'joins': [x.to_dict() for x in sorted(self.joins, key=lambda join: join.identifier())],
            'identifiers': sorted(self.identifiers),
            # The virtual_data_source and package are not exported to yaml
        }

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'PanoModel':
        return cls(
            model_name=inputs['model_name'],
            data_source=inputs['data_source'],
            fields=[PanoModelField.from_dict(x) for x in inputs.get('fields', [])],
            joins=[PanoModelJoin.from_dict(x) for x in inputs.get('joins', [])],
            identifiers=inputs.get('identifiers', []),
            virtual_data_source=inputs.get('virtual_data_source'),
            package=inputs.get('package'),
            file_name=inputs.get('file_name'),
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.model_name,
                self.data_source,
                tuple(self.fields),
                tuple(self.joins),
                tuple(self.identifiers),
                self.virtual_data_source,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PanoModel):
            return False

        return self.to_dict() == o.to_dict()


class PanoVirtualDataSource(Actionable):
    """Group collection of models into one data source."""

    API_VERSION = 'v1'

    dataset_slug: str
    display_name: str

    def __init__(self, *, dataset_slug: str, display_name: str, package: Optional[str] = None):
        self.dataset_slug = dataset_slug
        self.display_name = display_name
        self.package = package

    @property
    def id(self) -> Tuple[str]:
        return (self.dataset_slug,)

    def to_dict(self) -> Dict[str, Any]:
        return {
            'api_version': self.API_VERSION,
            'dataset_slug': self.dataset_slug,
            'display_name': self.display_name,
            # The package is not exported to yaml
        }

    @classmethod
    def from_dict(cls, inputs: Dict[str, Any]) -> 'PanoVirtualDataSource':
        return cls(
            dataset_slug=inputs['dataset_slug'], display_name=inputs['display_name'], package=inputs.get('package')
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.dataset_slug,
                self.display_name,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, PanoVirtualDataSource):
            return False

        return self.to_dict() == o.to_dict()



from unittest.mock import MagicMock

import pytest

from panoramic.cli.clients import VersionedClient


def test_versioned_client_with_session():
    mock_session = MagicMock()

    class TestClient(VersionedClient):
        def __init__(self):
            self.session = mock_session

    TestClient()

    assert mock_session.session.headers['User-Agent'] is not None


def test_versioned_client_with_no_session():
    class TestClient(VersionedClient):
        pass

    with pytest.raises(AssertionError):
        TestClient()



import tempfile
from pathlib import Path

import pytest
import yaml

from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.paths import Paths


@pytest.fixture(autouse=True)
def remove_client_creds_env(monkeypatch):
    monkeypatch.delenv('PANO_CLIENT_ID', raising=False)
    monkeypatch.delenv('PANO_CLIENT_SECRET', raising=False)


def test_env_vars(monkeypatch):
    with tempfile.TemporaryDirectory() as tmpdirname:
        monkeypatch.setattr(Path, 'home', lambda: Path(tmpdirname))
        Paths.config_dir().mkdir()

        with Paths.config_file().open('w') as f:
            f.write(yaml.dump(dict(client_id='some_random_id', client_secret='some_random_secret')))

        monkeypatch.setenv('PANO_CLIENT_ID', 'test_client_id')
        monkeypatch.setenv('PANO_CLIENT_SECRET', 'test_client_secret')

        # Use __wrapped__ to avoid cache
        assert get_client_id.__wrapped__() == 'test_client_id'
        assert get_client_secret.__wrapped__() == 'test_client_secret'


def test_config_file(monkeypatch):
    with tempfile.TemporaryDirectory() as tmpdirname:
        monkeypatch.setattr(Path, 'home', lambda: Path(tmpdirname))
        Paths.config_dir().mkdir()

        with Paths.config_file().open('w') as f:
            f.write(yaml.dump(dict(client_id='some_random_id', client_secret='some_random_secret')))

        # Use __wrapped__ to avoid cache
        assert get_client_id.__wrapped__() == 'some_random_id'
        assert get_client_secret.__wrapped__() == 'some_random_secret'



from unittest.mock import Mock, call, patch, sentinel

import pytest
from requests.exceptions import HTTPError

from panoramic.cli.errors import (
    MissingFieldFileError,
    ScanException,
    SourceNotFoundException,
)
from panoramic.cli.metadata.client import TERMINAL_STATES, JobState
from panoramic.cli.scan import Scanner, scan_fields_for_errors


@patch('panoramic.cli.scan.MetadataClient')
def test_scan_columns_completed(mock_client):
    mock_client.return_value.create_get_columns_job.return_value = 'test-job-id'
    mock_client.return_value.wait_for_terminal_state.return_value = JobState.COMPLETED
    mock_client.return_value.collect_results.return_value = [sentinel.value1, sentinel.value2]

    assert list(Scanner('test-company', 'test-source').scan_columns(table_filter='table-name')) == [
        sentinel.value1,
        sentinel.value2,
    ]


@pytest.mark.parametrize('final_state', TERMINAL_STATES - {JobState.COMPLETED})
@patch('panoramic.cli.scan.MetadataClient')
def test_scan_columns_non_completed(mock_client, final_state):
    mock_client.return_value.create_get_columns_job.return_value = 'test-job-id'
    mock_client.return_value.wait_for_terminal_state.return_value = final_state

    # TODO: add proper exception
    with pytest.raises(Exception):
        list(Scanner('test-company', 'test-source').scan_columns(table_filter='table-name'))


@patch('panoramic.cli.scan.MetadataClient')
def test_scan_columns_not_found(mock_client):
    mock_client.return_value.create_get_columns_job.side_effect = HTTPError(response=Mock(status_code=404))

    with pytest.raises(SourceNotFoundException):
        list(Scanner('test-company', 'test-source').scan_columns(table_filter='table-name', timeout=1))


@patch('panoramic.cli.scan.MetadataClient')
def test_scan_columns_generic_error(mock_client):
    mock_client.return_value.create_get_columns_job.side_effect = HTTPError(response=Mock(status_code=500))

    with pytest.raises(ScanException):
        list(Scanner('test-company', 'test-source').scan_columns(table_filter='table-name', timeout=1))


@patch('panoramic.cli.scan.MetadataClient')
def test_scan_tables_completed(mock_client):
    mock_client.return_value.create_get_tables_job.return_value = 'test-job-id'
    mock_client.return_value.wait_for_terminal_state.return_value = JobState.COMPLETED
    mock_client.return_value.collect_results.return_value = [sentinel.value1, sentinel.value2]

    assert list(Scanner('test-company', 'test-source').scan_tables(table_filter='table-name')) == [
        sentinel.value1,
        sentinel.value2,
    ]


@pytest.mark.parametrize('final_state', TERMINAL_STATES - {JobState.COMPLETED})
@patch('panoramic.cli.scan.MetadataClient')
def test_scan_tables_non_completed(mock_client, final_state):
    mock_client.return_value.create_get_tables_job.return_value = 'test-job-id'
    mock_client.return_value.get_job_status.return_value = final_state

    # TODO: add proper exception
    with pytest.raises(Exception):
        list(Scanner('test-company', 'test-source').scan_tables(table_filter='table-name'))


@patch('panoramic.cli.scan.MetadataClient')
def test_scan_tables_not_found(mock_client):
    mock_client.return_value.create_get_tables_job.side_effect = HTTPError(response=Mock(status_code=404))

    with pytest.raises(SourceNotFoundException):
        list(Scanner('test-company', 'test-source').scan_tables(table_filter='table-name', timeout=1))


@patch('panoramic.cli.scan.MetadataClient')
def test_scan_tables_generic_error(mock_client):
    mock_client.return_value.create_get_tables_job.side_effect = HTTPError(response=Mock(status_code=500))

    with pytest.raises(ScanException):
        list(Scanner('test-company', 'test-source').scan_tables(table_filter='table-name', timeout=1))


@patch('panoramic.cli.scan.get_company_slug', return_value='test-company')
@patch('panoramic.cli.scan.map_error_to_field')
@patch('panoramic.cli.scan.map_column_to_field')
@patch('panoramic.cli.scan.Scanner')
def test_scan_fields_for_errors(
    mock_scanner,
    mock_map_column_to_field,
    mock_map_error_to_field,
    _,
):
    errors = [
        # scanned
        MissingFieldFileError(
            field_slug='test_slug_1',
            dataset_slug='test_dataset',
            data_source='db1.schema.table1',
            data_reference='"TEST_COLUMN"',
            identifier=False,
        ),
        # not scanned
        MissingFieldFileError(
            field_slug='test_slug_2',
            dataset_slug='test_dataset',
            data_source='db2.schema.table1',
            data_reference='10 + 5',
            identifier=False,
        ),
    ]

    mock_scanner_db1 = Mock()
    mock_scanner_db1.scan_columns.return_value = [
        {'data_reference': '"TEST_COLUMN"'},
        {'data_reference': '"SOME_OTHER_COLUMN"'},
    ]
    mock_scanner_db2 = Mock()
    mock_scanner_db2.scan_columns.return_value = [{'data_reference': '"SOME_OTHER_COLUMN"'}]

    mock_scanner.side_effect = lambda _, conn: mock_scanner_db1 if conn == 'db1' else mock_scanner_db2

    mock_map_column_to_field.return_value = sentinel.scanned_field
    mock_map_error_to_field.return_value = sentinel.non_scanned_field

    fields = scan_fields_for_errors(errors)

    assert mock_scanner_db2.scan_columns.mock_calls == [call(table_filter='schema.table1')]
    assert mock_scanner_db1.scan_columns.mock_calls == [call(table_filter='schema.table1')]
    assert set(fields) == {sentinel.scanned_field, sentinel.non_scanned_field}



import pytest

from panoramic.cli.model.client import Model, ModelField
from panoramic.cli.model_mapper import (
    map_columns_to_model,
    map_model_from_local,
    map_model_from_remote,
)
from panoramic.cli.pano_model import PanoModel, PanoModelField


@pytest.fixture
def local_model_fixture():
    yield PanoModel(
        model_name='source_schema1_table',
        data_source='source.schema1.table',
        fields=[
            PanoModelField(
                data_reference='id',
                field_map=['id'],
            ),
            PanoModelField(
                data_reference='value',
                field_map=['value'],
            ),
        ],
        joins=[],
        identifiers=[],
    )


@pytest.fixture
def remote_model_fixture():
    yield Model(
        model_name="source_schema1_table",
        data_source="source.schema1.table",
        fields=[
            ModelField(
                field_map=["id"],
                data_reference="id",
            ),
            ModelField(
                field_map=["value"],
                data_reference="value",
            ),
        ],
        joins=[],
        identifiers=[],
        visibility='available',
    )


def test_from_remote_to_local(remote_model_fixture, local_model_fixture):
    expected_local_model = local_model_fixture.to_dict()

    local_model = map_model_from_remote(remote_model_fixture)

    assert local_model.to_dict() == expected_local_model


def test_from_local_to_remote(local_model_fixture, remote_model_fixture):
    expected_remote_model = remote_model_fixture.to_dict()

    remote_model = map_model_from_local(local_model_fixture)

    assert remote_model.to_dict() == expected_remote_model


def test_map_columns_to_model():
    expected = [
        {
            'api_version': 'v1',
            'model_name': 'sourceschema1table1',
            'data_source': 'source.schema1.table1',
            'fields': [
                {
                    'field_map': ['id'],
                    'data_reference': 'id',
                },
                {
                    'field_map': ['value'],
                    'data_reference': 'value',
                },
            ],
            'identifiers': [],
            'joins': [],
        }
    ]
    output = [
        item.to_dict()
        for item in map_columns_to_model(
            [
                {
                    'field_map': ['id'],
                    'data_reference': 'id',
                    'model_name': 'sourceschema1table1',
                    'data_source': 'source.schema1.table1',
                },
                {
                    'field_map': ['value'],
                    'data_reference': 'value',
                    'model_name': 'sourceschema1table1',
                    'data_source': 'source.schema1.table1',
                },
            ],
        )
    ]

    assert output == expected



from typing import Any, Dict, Optional, Tuple

import click
import hologram
import yaml
from dbt.adapters.factory import (
    Credentials,
    get_adapter_class_by_name,
    get_config_class_by_name,
    load_plugin,
)
from dbt.exceptions import FailedToConnectException, RuntimeException

from panoramic.cli.config.storage import read_config, update_config
from panoramic.cli.errors import (
    ConnectionAlreadyExistsException,
    ConnectionCreateException,
    ConnectionNotFound,
    ConnectionUpdateException,
)
from panoramic.cli.paths import Paths
from panoramic.cli.print import echo_info

CONNECTION_KEYS = [
    'type',
    'user',
    'host',
    'port',
    'password',
    'database',
    'schema',
    'warehouse',
    'account',
    'project',
    'key_file',
]


def create_connection_command(
    name: str,
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    password_stdin: bool,
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
    no_test: bool,
) -> None:
    """CLI command. Create new connection."""
    connections = Connections.load()
    if name in connections:
        raise ConnectionAlreadyExistsException(name)

    if password_stdin:
        password = click.prompt('Enter password: ', hide_input=True, type=str)

    new_connection: Dict[str, Any] = {}
    _update_connection_from_args(
        new_connection,
        type=type,
        user=user,
        password=password,
        host=host,
        port=port,
        database=database,
        schema=schema,
        warehouse=warehouse,
        account=account,
        project=project,
        key_file=key_file,
    )

    credentials, error = get_dialect_credentials(new_connection)
    if error is not None:
        raise ConnectionCreateException(error)

    if not no_test:
        ok, error = Connections.test(credentials)
        if not ok:
            raise ConnectionCreateException(error)

    # Empty values are important for DBT credentials verification, but we don't need to store empty values in config.
    connections[name] = {}
    for key, value in new_connection.items():
        if value != '':
            connections[name][key] = value

    Connections.save(connections)
    echo_info('Connection was successfully created!')


def update_connection_command(
    name: str,
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    password_stdin: bool,
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
    no_test: bool,
) -> None:
    """CLI command. Update specific connection."""
    connections = Connections.load()
    if name not in connections:
        raise ConnectionNotFound(name)

    if password_stdin:
        password = click.prompt('Enter password: ', hide_input=True, type=str)

    new_connection = connections[name].copy()
    _update_connection_from_args(
        new_connection,
        type=type,
        user=user,
        password=password,
        host=host,
        port=port,
        database=database,
        schema=schema,
        warehouse=warehouse,
        account=account,
        project=project,
        key_file=key_file,
    )

    credentials, error = get_dialect_credentials(new_connection)
    if error is not None:
        raise ConnectionUpdateException(error)

    if not no_test:
        ok, error = Connections.test(credentials)
        if not ok:
            raise ConnectionUpdateException(error)

    # Empty values are important for DBT credentials verification, but we don't need to store empty values in config.
    for key, value in new_connection.items():
        if value != '':
            connections[name][key] = value

    Connections.save(connections)
    echo_info('Connection was successfully updated!')


def list_connections_command(show_password: bool) -> None:
    """CLI command. List all connections."""
    connections = Connections.load()
    if not connections:
        config_file = Paths.config_file()
        echo_info(
            f'No connections found.\n'
            f'Use "pano connection create" to create connection or edit "{config_file}" file.'
        )
        exit(0)

    if not show_password:
        for conn in connections.values():
            conn['password'] = '*****'
    echo_info(yaml.dump(connections))


def remove_connection_command(name: str) -> None:
    """CLI command. Remove existing connection."""
    connections = Connections.load()
    if name not in connections:
        raise ConnectionNotFound(name)

    del connections[name]
    Connections.save(connections)


def test_connections_command(name: Optional[str] = None) -> None:
    """CLI command. Test connections by trying to connect to the database.
    Optionally you can specify name for specific connection to that only that."""
    connections = Connections.load()
    if name is not None:
        if name not in connections:
            raise ConnectionNotFound(name)
        # Filter specified connection by name
        connections = {name: connections[name]}

    for name, connection in connections.items():
        # Fill with default values because DBT requires some fields we don't.
        for key in CONNECTION_KEYS:
            if key not in connection:
                connection[key] = ''

        credentials, error = get_dialect_credentials(connection)
        if error is not None:
            echo_info(f'{name}... FAIL: {error}')
            continue

        ok, error = Connections.test(credentials)
        if ok:
            echo_info(f'{name}... OK')
        else:
            echo_info(f'{name}... FAIL: {error}')


class Connections:
    @classmethod
    def get_by_name(cls, name: str) -> Dict[str, str]:
        connections = cls.load()
        return connections[name]

    @staticmethod
    def save(data: Dict[str, Any]) -> None:
        """Save connections YAML."""
        update_config('connections', data)

    @staticmethod
    def load() -> Dict[str, Any]:
        """Load connections YAML."""
        return read_config('connections')

    @classmethod
    def test(cls, credentials) -> Tuple[bool, str]:
        """Test connection string by connecting using DBT."""
        # Create dialect specific configuration
        adapter_config_cls = get_config_class_by_name(credentials.type)
        adapter_config = adapter_config_cls()
        adapter_config.credentials = credentials  # type: ignore

        # Create dialect specific adapter that handles connections
        adapter_cls = get_adapter_class_by_name(credentials.type)
        adapter = adapter_cls(adapter_config)  # type: ignore

        # Create dialect specific connection
        connection = adapter.acquire_connection()  # type: ignore
        try:
            # This will try to connect to remote database
            connection.handle
        except FailedToConnectException as e:
            return False, str(e)
        return True, ''


def _update_connection_from_args(
    connection: Dict[str, Any],
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
):
    if type is not None:
        connection['type'] = type
    if user is not None:
        connection['user'] = user
    if password is not None:
        connection['password'] = password
    if host is not None:
        connection['host'] = host
    if port is not None:
        connection['port'] = port
    if database is not None:
        connection['database'] = database
    if schema is not None:
        connection['schema'] = schema
    if warehouse is not None:
        connection['warehouse'] = warehouse
    if account is not None:
        connection['account'] = account
    if project is not None:
        connection['project'] = project
    if key_file is not None:
        connection['key_file'] = key_file

    # Fill with default values because DBT requires some fields we don't.
    for key in CONNECTION_KEYS:
        if key not in connection:
            connection[key] = ''
    return connection


def get_dialect_credentials(connection: Dict[str, Any]) -> (Tuple[Optional[Credentials], Optional[str]]):
    """Use DBT lib to create dialect specific credentials."""
    connection = connection.copy()
    type_ = connection.pop('type')

    try:
        plugin_cls = load_plugin(type_)
    except RuntimeException as e:
        return None, str(e)

    try:
        credentials = plugin_cls.from_dict(connection)
    except hologram.ValidationError as e:
        return None, e.message  # noqa: B306
    return credentials, None



import logging
from typing import Any, Dict, List, Optional
from urllib.parse import urljoin

from panoramic.auth import OAuth2Client
from panoramic.cli.clients import VersionedClient
from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.config.model import get_base_url

logger = logging.getLogger(__name__)


class ModelField:

    field_map: List[str]
    data_reference: str

    def __init__(
        self,
        *,
        field_map: List[str],
        data_reference: str,
    ):
        self.field_map = field_map
        self.data_reference = data_reference

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'ModelField':
        return cls(
            field_map=data['field_map'],
            data_reference=data['data_reference'],
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            'field_map': self.field_map,
            'data_reference': self.data_reference,
        }

    def __hash__(self) -> int:
        return hash(
            (
                tuple(self.field_map),
                self.data_reference,
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ModelField):
            return False

        return self.to_dict() == o.to_dict()


class ModelJoin:

    to_model: str
    join_type: str
    relationship: str
    fields: List[str]

    def __init__(self, *, to_model: str, join_type: str, relationship: str, fields: List[str]):
        self.to_model = to_model
        self.join_type = join_type
        self.relationship = relationship
        self.fields = fields

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> 'ModelJoin':
        return cls(
            to_model=data['to_model'],
            join_type=data['join_type'],
            relationship=data['relationship'],
            fields=data['fields'],
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            'to_model': self.to_model,
            'join_type': self.join_type,
            'relationship': self.relationship,
            'fields': self.fields,
        }

    def __hash__(self) -> int:
        return hash(
            (
                self.to_model,
                self.join_type,
                self.relationship,
                tuple(self.fields),
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, ModelJoin):
            return False

        return self.to_dict() == o.to_dict()


class Model:

    model_name: str
    data_source: str
    fields: List[ModelField]
    joins: List[ModelJoin]
    visibility: str
    virtual_data_source: Optional[str]
    identifiers: List[str]

    def __init__(
        self,
        *,
        model_name: str,
        data_source: str,
        fields: List[ModelField],
        joins: List[ModelJoin],
        identifiers: List[str],
        visibility: str,
        virtual_data_source: Optional[str] = None,
    ):
        self.model_name = model_name
        self.data_source = data_source
        self.fields = fields
        self.joins = joins
        self.identifiers = identifiers
        self.visibility = visibility
        self.virtual_data_source = virtual_data_source

    @classmethod
    def from_dict(cls, data: Dict[str, Any], **kwargs) -> 'Model':
        return cls(
            model_name=data['model_name'],
            data_source=data['data_source'],
            fields=[ModelField.from_dict(a) for a in data['fields']],
            joins=[ModelJoin.from_dict(d) for d in data['joins']],
            identifiers=data['identifiers'],
            visibility=data['visibility'],
            **kwargs,
        )

    def to_dict(self) -> Dict[str, Any]:
        return {
            'model_name': self.model_name,
            'data_source': self.data_source,
            'fields': [a.to_dict() for a in self.fields],
            'joins': [j.to_dict() for j in self.joins],
            'identifiers': self.identifiers,
            'visibility': self.visibility,
        }

    def __hash__(self) -> int:
        return hash(
            (
                self.model_name,
                self.data_source,
                tuple(self.fields),
                tuple(self.joins),
                self.visibility,
                self.virtual_data_source,
                tuple(self.identifiers),
            )
        )

    def __eq__(self, o: object) -> bool:
        if not isinstance(o, Model):
            return False

        return self.to_dict() == o.to_dict()


class ModelClient(OAuth2Client, VersionedClient):

    """Model HTTP API client."""

    base_url: str

    def __init__(
        self,
        base_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        base_url = base_url if base_url is not None else get_base_url()
        client_id = client_id if client_id is not None else get_client_id()
        client_secret = client_secret if client_secret is not None else get_client_secret()

        self.base_url = base_url
        super().__init__(client_id, client_secret)

    def delete_model(self, data_source: str, company_slug: str, name: str):
        """Delete model with a given name."""
        url = urljoin(self.base_url, name)
        logger.debug(f'Deleting model with name: {name}')
        params = {'virtual_data_source': data_source, 'company_slug': company_slug}
        response = self.session.delete(url, params=params, timeout=30)
        response.raise_for_status()

    def upsert_model(self, data_source: str, company_slug: str, model: Model):
        """Add or update given model."""
        logger.debug(f'Upserting model with name: {model.model_name}')
        params = {'virtual_data_source': data_source, 'company_slug': company_slug}
        response = self.session.put(self.base_url, json=model.to_dict(), params=params, timeout=30)
        response.raise_for_status()

    def get_model_names(self, data_source: str, company_slug: str, offset: int = 0, limit: int = 100) -> List[str]:
        """Retrieve names of all models in a given source."""
        logger.debug(f'Listing names of models for source: {data_source}')
        url = urljoin(self.base_url, 'model-name')
        params = {'virtual_data_source': data_source, 'company_slug': company_slug, 'offset': offset, 'limit': limit}
        response = self.session.get(url, params=params, timeout=30)
        response.raise_for_status()
        return response.json()['data']

    def get_models(self, data_source: str, company_slug: str, offset: int = 0, limit: int = 100) -> List[Model]:
        """Retrieve all models in a given source."""
        logger.debug(f'Listing models for source: {data_source}')
        params = {'virtual_data_source': data_source, 'company_slug': company_slug, 'offset': offset, 'limit': limit}
        response = self.session.get(
            self.base_url,
            params=params,
        )
        response.raise_for_status()
        return [Model.from_dict(d, virtual_data_source=data_source) for d in response.json()['data']]



import json
import logging
from pathlib import Path
from typing import IO, Any, Dict, List, Optional, Union

import yaml

from panoramic.cli.errors import FileMissingError, InvalidYamlFile

logger = logging.getLogger(__name__)


def load_yaml(text: Union[bytes, IO[bytes], str, IO[str]]) -> Any:
    """Load YAML from stream."""
    return yaml.safe_load(text)


def dump_yaml(data: Any, stream: Optional[IO[str]] = None) -> Optional[str]:
    """Dump YAML to stream or return as string."""
    if stream is not None:
        return yaml.safe_dump(data, stream, default_flow_style=False)
    else:
        return yaml.safe_dump(data, default_flow_style=False)


def ensure_dir(abs_filepath: Path):
    """
    Ensure parent directory exists.
    """
    path_obj = abs_filepath.parent
    path_obj.mkdir(parents=True, exist_ok=True)


def write_yaml(abs_filepath: Path, yaml_dict: Dict[str, Any]):
    """
    Writes yaml dict to path
    """
    logger.debug(f'Write yaml {abs_filepath}')
    ensure_dir(abs_filepath)
    with open(abs_filepath, 'w') as f:
        dump_yaml(yaml_dict, f)


def read_yaml(path: Path) -> Dict[str, Any]:
    """
    Reads yaml dict from path
    """
    logger.debug(f'Read yaml {path}')
    try:
        with open(path, 'r') as f:
            return load_yaml(f)
    except FileNotFoundError:
        raise FileMissingError(path=path)
    except yaml.MarkedYAMLError as e:
        raise InvalidYamlFile(path=path, error=e)


def append_json_line(abs_filepath: Path, json_dict: Dict[str, Any]) -> None:
    """
    Writes json dict to path as single line. Useful to create JSON lines file. https://jsonlines.org/
    """
    logger.debug(f'Write json {abs_filepath}')
    ensure_dir(abs_filepath)
    with open(abs_filepath, 'a') as f:
        data = json.dumps(json_dict)
        f.write(data + '\n')


def read_json_lines(path: Path) -> List[Dict[str, Any]]:
    """
    Reads json lines from path. https://jsonlines.org/
    """
    logger.debug(f'Read json {path}')
    try:
        json_lines = []
        with open(path, 'r') as f:
            for line in f.readlines():
                json_lines.append(json.loads(line))
        return json_lines
    except FileNotFoundError:
        raise FileMissingError(path=path)


def truncate_file(path: Path) -> None:
    """
    Opening file in write mode will remove files content.
    """
    open(path, 'w').close()


def delete_file(abs_filepath: Path):
    """Delete file at given path."""
    # TODO: Consider warning - wanted to delete model but not found
    if abs_filepath.exists():
        abs_filepath.unlink()



from unittest.mock import Mock, call, patch

from panoramic.cli.local.writer import FileWriter
from panoramic.cli.pano_model import PanoField, PanoModel
from panoramic.cli.paths import FileExtension, PresetFileName, SystemDirectory


@patch('panoramic.cli.local.writer.write_yaml')
def test_writer_write_data_source_package_none(mock_write_yaml, tmp_path):
    mock_ds = Mock(dataset_slug='test_dataset')

    FileWriter(cwd=tmp_path).write_data_source(mock_ds)

    assert mock_write_yaml.mock_calls == [
        call(tmp_path / 'test_dataset' / PresetFileName.DATASET_YAML.value, mock_ds.to_dict.return_value)
    ]


@patch('panoramic.cli.local.writer.write_yaml')
def test_writer_write_data_source_package_set(mock_write_yaml, tmp_path):
    mock_ds = Mock(dataset_slug='test_dataset')

    FileWriter(cwd=tmp_path).write_data_source(mock_ds, package='my_package')

    assert mock_write_yaml.mock_calls == [
        call(tmp_path / 'my_package' / PresetFileName.DATASET_YAML.value, mock_ds.to_dict.return_value)
    ]


@patch('panoramic.cli.local.writer.delete_file')
def test_writer_delete_data_source(mock_delete_file, tmp_path):
    mock_ds = Mock(package='test_dataset')

    FileWriter(cwd=tmp_path).delete_data_source(mock_ds)

    assert mock_delete_file.mock_calls == [call(tmp_path / 'test_dataset' / PresetFileName.DATASET_YAML.value)]


@patch('panoramic.cli.local.writer.write_yaml')
def test_writer_write_field(mock_write_yaml, tmp_path):
    mock_field_company_scoped = Mock(spec=PanoField, slug='slug', data_source=None)
    mock_field_vds_scoped = Mock(spec=PanoField, slug='slug', data_source='test_dataset')

    FileWriter(cwd=tmp_path).write_field(mock_field_company_scoped)
    FileWriter(cwd=tmp_path).write_field(mock_field_vds_scoped)

    assert mock_write_yaml.mock_calls == [
        call(
            tmp_path
            / SystemDirectory.FIELDS.value
            / f'{mock_field_company_scoped.slug}{FileExtension.FIELD_YAML.value}',
            mock_field_company_scoped.to_dict.return_value,
        ),
        call(
            tmp_path
            / 'test_dataset'
            / SystemDirectory.FIELDS.value
            / f'{mock_field_vds_scoped.slug}{FileExtension.FIELD_YAML.value}',
            mock_field_vds_scoped.to_dict.return_value,
        ),
    ]


@patch('panoramic.cli.local.writer.delete_file')
def test_writer_delete_field(mock_delete_file, tmp_path):
    file_name = f'slug{FileExtension.FIELD_YAML.value}'
    mock_field_company_scoped = Mock(spec=PanoField, slug='slug', package=None, file_name=file_name)
    mock_field_vds_scoped = Mock(spec=PanoField, slug='slug', package='test_dataset', file_name=file_name)

    FileWriter(cwd=tmp_path).delete_field(mock_field_company_scoped)
    FileWriter(cwd=tmp_path).delete_field(mock_field_vds_scoped)

    assert mock_delete_file.mock_calls == [
        call(tmp_path / SystemDirectory.FIELDS.value / file_name),
        call(tmp_path / 'test_dataset' / SystemDirectory.FIELDS.value / file_name),
    ]


@patch('panoramic.cli.local.writer.Paths')
@patch('panoramic.cli.local.writer.write_yaml')
def test_writer_write_scanned_model(mock_write_yaml, mock_paths, tmp_path):
    mock_paths.scanned_dir.return_value = tmp_path
    mock_model = Mock(spec=PanoModel, model_name='model')

    FileWriter(cwd=tmp_path).write_scanned_model(mock_model)

    assert mock_write_yaml.mock_calls == [
        call(tmp_path / f'model{FileExtension.MODEL_YAML.value}', mock_model.to_dict())
    ]


@patch('panoramic.cli.local.writer.Paths')
@patch('panoramic.cli.local.writer.write_yaml')
def test_writer_write_scanned_field(mock_write_yaml, mock_paths, tmp_path):
    mock_paths.scanned_fields_dir.return_value = tmp_path
    mock_field = Mock(spec=PanoField, slug='slug')

    FileWriter(cwd=tmp_path).write_scanned_field(mock_field)

    assert mock_write_yaml.mock_calls == [
        call(tmp_path / f'slug{FileExtension.FIELD_YAML.value}', mock_field.to_dict())
    ]


@patch('panoramic.cli.local.writer.Paths')
def test_writer_write_empty_model(mock_paths, tmp_path):
    mock_model = Mock(spec=PanoModel, model_name='model')

    FileWriter(cwd=tmp_path).write_empty_model(mock_model)

    assert (mock_paths.scanned_dir.return_value / f'model{FileExtension.MODEL_YAML.value}').touch.mock_calls == [call()]


@patch('panoramic.cli.local.writer.Paths')
def test_writer_write_empty_field(mock_paths, tmp_path):
    mock_field = Mock(spec=PanoField, slug='slug')

    FileWriter(cwd=tmp_path).write_empty_field(mock_field)

    assert (mock_paths.scanned_fields_dir.return_value / f'model{FileExtension.FIELD_YAML.value}').touch.mock_calls == [
        call()
    ]


@patch('panoramic.cli.local.writer.delete_file')
def test_writer_delete_model(mock_delete_file, tmp_path):
    file_name = f'model{FileExtension.FIELD_YAML.value}'
    mock_model = Mock(spec=PanoModel, model_name='model', package='package', file_name=file_name)

    FileWriter(cwd=tmp_path).delete_model(mock_model)

    assert mock_delete_file.mock_calls == [
        call(tmp_path / 'package' / file_name),
    ]


@patch('panoramic.cli.local.writer.write_yaml')
def test_writer_write_model(mock_write_yaml, tmp_path):
    mock_model = Mock(spec=PanoModel, model_name='model', virtual_data_source='test_dataset')

    FileWriter(cwd=tmp_path).write_model(mock_model)

    assert mock_write_yaml.mock_calls == [
        call(tmp_path / 'test_dataset' / f'model{FileExtension.MODEL_YAML.value}', mock_model.to_dict()),
    ]



import pytest
import responses

from panoramic.cli.physical_data_source.client import PhysicalDataSourceClient


@pytest.fixture(autouse=True)
def mock_token_url(monkeypatch):
    monkeypatch.setenv('PANORAMIC_AUTH_TOKEN_URL', 'https://token')


@responses.activate
def test_get_sources():
    responses.add(responses.POST, 'https://token/', json={'access_token': '123123'})
    responses.add(
        responses.GET,
        'https://diesel/source/?company_slug=company_name',
        json={
            'data': [
                # ... sth should be here
            ]
        },
    )

    client = PhysicalDataSourceClient(
        base_url='https://diesel/source/', client_id='client-id', client_secret='client-secret'
    )
    client.get_sources('company_name')

    assert len(responses.calls) == 2



import logging
import time
from enum import Enum
from typing import Any, Dict, Iterable, List, Optional
from urllib.parse import urljoin

from panoramic.auth import OAuth2Client
from panoramic.cli.clients import VersionedClient
from panoramic.cli.config.auth import get_client_id, get_client_secret
from panoramic.cli.config.metadata import get_base_url
from panoramic.cli.errors import TimeoutException

logger = logging.getLogger(__name__)


class JobState(Enum):

    NOT_SUBMITTED = 'NOT_SUBMITTED'
    STARTING = 'STARTING'
    RUNNING = 'RUNNING'
    COMPLETED = 'COMPLETED'
    CANCELED = 'CANCELED'
    FAILED = 'FAILED'
    CANCELLATION_REQUESTED = 'CANCELLATION_REQUESTED'
    ENQUEUED = 'ENQUEUED'
    PLANNING = 'PLANNING'
    PENDING = 'PENDING'
    METADATA_RETRIEVAL = 'METADATA_RETRIEVAL'
    QUEUED = 'QUEUED'
    ENGINE_START = 'ENGINE_START'
    EXECUTION_PLANNING = 'EXECUTION_PLANNING'
    INVALID_STATE = 'INVALID_STATE'


TERMINAL_STATES = {JobState.COMPLETED, JobState.CANCELED, JobState.FAILED}


class MetadataClient(OAuth2Client, VersionedClient):

    """Metadata HTTP API client."""

    base_url: str

    def __init__(
        self,
        base_url: Optional[str] = None,
        client_id: Optional[str] = None,
        client_secret: Optional[str] = None,
    ):
        base_url = base_url if base_url is not None else get_base_url()
        client_id = client_id if client_id is not None else get_client_id()
        client_secret = client_secret if client_secret is not None else get_client_secret()

        super().__init__(client_id, client_secret)
        self.base_url = base_url

    def create_refresh_job(self, company_slug: str, source_id: str, table_name: str):
        """Starts async "refresh metadata" job and return job id."""
        url = urljoin(self.base_url, f'{source_id}/refresh')
        params = {'table_name': table_name, 'company_slug': company_slug}
        logger.debug(f'Refreshing table for source {source_id} and name: {table_name}')
        response = self.session.post(url, params=params, timeout=30)
        response.raise_for_status()
        return response.json()['data']['job_id']

    def create_get_tables_job(self, company_slug: str, source_id: str, table_filter: Optional[str]) -> str:
        """Starts async "get tables" job and return job id."""
        url = urljoin(self.base_url, f'{source_id}/tables')
        params = (
            {'company_slug': company_slug}
            if table_filter is None
            else {'company_slug': company_slug, 'table_filter': table_filter}
        )
        logger.debug(f'Requesting tables for source {source_id} and filter: {table_filter}')
        response = self.session.post(url, params=params, timeout=30)
        response.raise_for_status()
        return response.json()['data']['job_id']

    def create_get_columns_job(self, company_slug: str, source_id: str, table_filter: Optional[str]) -> str:
        """Starts async "get columns" job and return job id."""
        url = urljoin(self.base_url, f'{source_id}/columns')
        params = (
            {'company_slug': company_slug}
            if table_filter is None
            else {'company_slug': company_slug, 'table_filter': table_filter}
        )
        logger.debug(f'Requesting columns for source {source_id} and filter: {table_filter}')
        response = self.session.post(url, params=params, timeout=30)
        response.raise_for_status()
        return response.json()['data']['job_id']

    def get_job_status(self, job_id: str) -> JobState:
        """Get status of an async job."""
        url = urljoin(self.base_url, f'job/{job_id}')
        logger.debug(f'Getting job status for job {job_id}')
        response = self.session.get(url, timeout=30)
        response.raise_for_status()
        return JobState(response.json()['data']['job_status'])

    def get_job_results(self, job_id: str, offset: int = 0, limit: int = 500) -> List[Dict[str, Any]]:
        """Get results of an async job."""
        url = urljoin(self.base_url, f'job/{job_id}/results')
        params = {'offset': offset, 'limit': limit}
        logger.debug(f'Getting job results for job {job_id}')
        response = self.session.get(url, params=params, timeout=30)
        response.raise_for_status()
        return response.json()['data']

    def wait_for_terminal_state(self, job_id: str, timeout: int = 60) -> JobState:
        """Wait for job to reach terminal state."""
        tick_time = 1
        while True:
            logger.debug(f'Getting status for job with id {job_id}')
            status = self.get_job_status(job_id)
            logger.debug(f'Got status {status} for job with id {job_id}')
            if status in TERMINAL_STATES:
                return status
            if timeout <= 0:
                raise TimeoutException(f'Timed out waiting for job {job_id} to complete')
            time.sleep(tick_time)
            timeout -= tick_time

    def collect_results(self, job_id: str, limit: int = 500) -> Iterable[Dict[str, Any]]:
        """Collect all results for a given job."""
        offset = 0

        while True:
            logger.debug(f'Fetching page number {offset // limit + 1} for job with id {job_id}')
            page = self.get_job_results(job_id, offset=offset, limit=limit)
            yield from page

            if len(page) < limit:
                # last page
                logger.debug(f'Finished fetching all results for job with id {job_id}')
                return

            offset += limit



import logging
import sys
import time
from collections import defaultdict
from typing import Optional

import click
from click.core import Command, Context
from dotenv import load_dotenv

from panoramic.cli.__version__ import __version__
from panoramic.cli.analytics import is_enabled, write_command_event
from panoramic.cli.command import configure_anonymous_analytics
from panoramic.cli.errors import (
    CompanyNotFoundException,
    SourceNotFoundException,
    ValidationError,
    ValidationErrorSeverity,
    handle_exception,
    handle_interrupt,
)
from panoramic.cli.paths import Paths
from panoramic.cli.print import echo_error, echo_errors, echo_info, echo_warnings


class CommandWithAnalytics(Command):
    """Execute command and write usage metrics to a file that is conditionally flushed to external system."""

    def invoke(self, ctx: Context):
        # This will make sure to display opt-out message for existing users that already run pano configure.
        if not is_enabled():
            configure_anonymous_analytics()

        # If command is a subcommand, combine its name with the command name.
        group = ''
        if ctx.parent is not None and ctx.parent.command != cli:  # type: ignore
            group = ctx.parent.command.name  # type: ignore

        start_time = time.time()
        try:
            result = super().invoke(ctx)
            write_command_event(self.name, group, start_time)
            return result
        except Exception as e:
            write_command_event(self.name, group, start_time, error=str(e))
            raise e


class ConfigAwareCommand(CommandWithAnalytics):
    """Perform config file validation before running command."""

    def invoke(self, ctx: Context):
        from panoramic.cli.validate import validate_config

        try:
            validate_config()
            return super().invoke(ctx)
        except ValidationError as e:
            echo_error(str(e))
            sys.exit(1)


class ConnectionAwareCommand(ConfigAwareCommand):
    def invoke(self, ctx: Context):
        try:
            return super().invoke(ctx)
        except Exception as e:
            echo_error(str(e))
            sys.exit(1)


class ContextAwareCommand(ConfigAwareCommand):
    """
    Perform config and context file validation before running command.

    Failure scenarios we handle:
      * Invalid config/context/local state - ValidationError
      * Company not found for user - CompanyNotFoundException
      * Source not found for company - SourceNotFoundException

    Any other error - we show stack trace.
    """

    def invoke(self, ctx: Context):
        from panoramic.cli.validate import validate_context

        try:
            validate_context()
            return super().invoke(ctx)
        except (ValidationError, SourceNotFoundException, CompanyNotFoundException) as e:
            echo_error(str(e))
            sys.exit(1)


class LocalStateAwareCommand(ContextAwareCommand):
    """Perform config, context, and local state files validation before running command."""

    def invoke(self, ctx: Context):
        from panoramic.cli.validate import validate_local_state

        errors_by_severity: defaultdict = defaultdict(list)
        for error in validate_local_state():
            errors_by_severity[error.severity].append(error)

        if len(errors_by_severity[ValidationErrorSeverity.WARNING]) > 0:
            echo_warnings(errors_by_severity[ValidationErrorSeverity.WARNING])
            echo_info('')

        if len(errors_by_severity[ValidationErrorSeverity.ERROR]) > 0:
            echo_errors(errors_by_severity[ValidationErrorSeverity.ERROR])
            sys.exit(1)

        return super().invoke(ctx)


@click.group(context_settings={'help_option_names': ["-h", "--help"]}, help='')
@click.option('--debug', is_flag=True, help='Enables debug mode')
@click.version_option(__version__)
@handle_exception
def cli(debug):
    """Run checks at the beginning of every command."""
    if debug:
        logger = logging.getLogger()
        logger.setLevel("DEBUG")

    load_dotenv(dotenv_path=Paths.dotenv_file())

    from panoramic.cli.supported_version import is_version_supported

    if not is_version_supported(__version__):
        sys.exit(1)


@cli.command(help='Scan models from source', cls=ContextAwareCommand)
@click.argument('source-id', type=str, required=True)
@click.option('--filter', '-f', type=str, help='Filter down what schemas to scan')
@click.option('--generate-identifiers', '-i', is_flag=True, help='Generate identifiers for models')
@click.option('--parallel', '-p', type=int, default=8, help='Parallelize metadata scan')
@handle_exception
@handle_interrupt
def scan(source_id: str, filter: Optional[str], parallel: int, generate_identifiers: bool):
    from panoramic.cli.command import scan as scan_command

    scan_command(source_id, filter, parallel, generate_identifiers)


@cli.command(help='Pull models from remote', cls=LocalStateAwareCommand)
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--diff', '-d', is_flag=True, help='Show the difference between local and remote state')
@handle_exception
def pull(yes: bool, target_dataset: str, diff: bool):
    from panoramic.cli.command import pull as pull_command

    pull_command(yes=yes, target_dataset=target_dataset, diff=diff)


@cli.command(help='Push models to remote', cls=LocalStateAwareCommand)
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--diff', '-d', is_flag=True, help='Show the difference between local and remote state')
@handle_exception
def push(yes: bool, target_dataset: str, diff: bool):
    from panoramic.cli.command import push as push_command

    push_command(yes=yes, target_dataset=target_dataset, diff=diff)


@cli.command(help='Configure pano CLI options')
@handle_exception
def configure():
    from panoramic.cli.command import configure as config_command

    config_command()


@cli.command(help='Initialize metadata repository', cls=ConfigAwareCommand)
@handle_exception
def init():
    from panoramic.cli.command import initialize

    initialize()


@cli.command(help='List available data connections', cls=ContextAwareCommand)
@handle_exception
def list_connections():
    from panoramic.cli.command import list_connections as list_connections_command

    list_connections_command()


@cli.command(help='List available companies', cls=ConfigAwareCommand)
@handle_exception
def list_companies():
    from panoramic.cli.command import list_companies as list_companies_command

    list_companies_command()


@cli.command(help='Validate local files', cls=CommandWithAnalytics)
@handle_exception
def validate():
    from panoramic.cli.command import validate as validate_command

    if not validate_command():
        sys.exit(1)


@cli.command(help='Detect joins under a dataset', cls=LocalStateAwareCommand)
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@click.option('--diff', '-d', is_flag=True, help='Show the difference between local and detected joins')
@click.option(
    '--overwrite', is_flag=True, default=False, help='Overwrite joins on local model files by suggestions from remote'
)
@handle_exception
def detect_joins(target_dataset: str, yes: bool, diff: bool, overwrite: bool):
    from panoramic.cli.command import detect_joins as detect_joins_command

    detect_joins_command(target_dataset=target_dataset, diff=diff, overwrite=overwrite, yes=yes)


@cli.group(name='field')
def field_cli():
    """Commands on local field files."""


@field_cli.command(help='Clean up fields not linked to any model', cls=ContextAwareCommand)
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@handle_exception
def cleanup(target_dataset: str, yes: bool):
    from panoramic.cli.command import delete_orphaned_fields as delete_orphans_command

    delete_orphans_command(target_dataset=target_dataset, yes=yes)


@field_cli.command(help='Scaffold fields defined in models', cls=ContextAwareCommand)
@click.option('--target-dataset', '-t', type=str, help='Target a specific dataset')
@click.option('--yes', '-y', is_flag=True, default=False, help='Automatically confirm all actions')
@handle_exception
def scaffold(target_dataset: str, yes: bool):
    from panoramic.cli.command import (
        scaffold_missing_fields as scaffold_missing_fields_command,
    )

    scaffold_missing_fields_command(target_dataset=target_dataset, yes=yes)


@cli.group()
def connection():
    """Connection subcommand for managing connections.

    \b
    All connections are stored in ~/.pano/config file.
    You can edit this file either manually or using provided commands.

    Expected YAML structure:

    \b
    name:
      type: postgres
      user: my_user
      password: <password>
      host: localhost
      port: 5432
      database: my_db

    Which effectively results in a connection string similar to: "postgres://my_user@<password>@localhost:5432/my_db".

    \b
    name: The name of the connection that will be used as reference to specify which connection to use.
    type: The type of data warehouse you are connecting to.
    user: The username used for database connection.
    password: The password used for database connection.
    host: The hostname (server) used for database connection.
    port: The port used for database connection.
    database: The database used for database connection.
    schema: The schema to build models into by default. Can be overridden by custom models.
    warehouse: The warehouse to use when building models. (Used by Snowflake)
    account: The account used for database connection. This will be something like cc123 or cc123.us-east-1 for your particular account. (Used by Snowflake)
    project: The project used for database connection. (Used by BigQuery)
    key_file: Keyfile path to Service Account JSON. (Used by BigQuery)
    """
    pass


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', type=str)
@click.option('--type', default=None, type=str, help='Type of the database. E.g. "postgres", "snowflake".')
@click.option('--user', default=None, type=str, help='Connection username.')
@click.option('--password', default=None, type=str, help='Connection password.')
@click.option('--password-stdin', default=False, is_flag=True, help='Read connection password from standard input.')
@click.option('--host', default=None, type=str, help='Connection hostname.')
@click.option('--port', default=None, type=int, help='Connection port.')
@click.option('--database', default=None, type=str, help='Connection database name.')
@click.option('--schema', default=None, type=str, help='Connection schema name.')
@click.option('--warehouse', default=None, type=str, help='Connection warehouse name. (Used by Snowflake)')
@click.option('--account', default=None, type=str, help='Connection account name. (Used by Snowflake)')
@click.option('--project', default=None, type=str, help='Connection project name. (Used by BigQuery)')
@click.option('--key-file', default=None, type=str, help='Keyfile path to Service Account JSON. (Used by BigQuery)')
@click.option('--no-test', default=False, is_flag=True, help='Do NOT try test the connection.')
def create(
    name: str,
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    password_stdin: bool,
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
    no_test: bool,
):
    """Add new connection.

    pano connection create postgres-prod --type postgres --user my_user \\
     --password-stdin --host localhost --port 5432 --database my_db
    """
    from panoramic.cli.connections import create_connection_command

    create_connection_command(
        name=name,
        type=type,
        user=user,
        host=host,
        port=port,
        password=password,
        password_stdin=password_stdin,
        database=database,
        schema=schema,
        warehouse=warehouse,
        account=account,
        project=project,
        key_file=key_file,
        no_test=no_test,
    )


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', type=str)
@click.option('--type', default=None, type=str, help='Type of the database. E.g. "postgres", "snowflake".')
@click.option('--user', default=None, type=str, help='Connection username.')
@click.option('--password', default=None, type=str, help='Connection password.')
@click.option('--password-stdin', default=False, is_flag=True, help='Read connection password from standard input.')
@click.option('--host', default=None, type=str, help='Connection hostname.')
@click.option('--port', default=None, type=int, help='Connection port.')
@click.option('--database', default=None, type=str, help='Connection database name.')
@click.option('--schema', default=None, type=str, help='Connection schema name.')
@click.option('--warehouse', default=None, type=str, help='Connection warehouse name. (Used by Snowflake)')
@click.option('--account', default=None, type=str, help='Connection account name. (Used by Snowflake)')
@click.option('--project', default=None, type=str, help='Connection project name. (Used by BigQuery)')
@click.option('--key-file', default=None, type=str, help='Keyfile path to Service Account JSON. (Used by BigQuery)')
@click.option('--no-test', default=False, is_flag=True, help='Do NOT try test the connection.')
def update(
    name: str,
    type: Optional[str],
    user: Optional[str],
    host: Optional[str],
    port: Optional[int],
    password: Optional[str],
    password_stdin: bool,
    database: Optional[str],
    schema: Optional[str],
    warehouse: Optional[str],
    account: Optional[str],
    project: Optional[str],
    key_file: Optional[str],
    no_test: bool,
):
    """Update existing connection.

    pano connection update postgres-prod --database my_new_prod_db
    """
    from panoramic.cli.connections import update_connection_command

    update_connection_command(
        name=name,
        type=type,
        user=user,
        host=host,
        port=port,
        password=password,
        password_stdin=password_stdin,
        database=database,
        schema=schema,
        warehouse=warehouse,
        account=account,
        project=project,
        key_file=key_file,
        no_test=no_test,
    )


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', type=str)
def remove(name: str):
    """Remove existing connection.

    pano connection remove postgres-prod
    """
    from panoramic.cli.connections import remove_connection_command

    remove_connection_command(name)


@connection.command(name='list', cls=ConnectionAwareCommand)
@click.option('--show-password', default=False, is_flag=True, help='Show passwords.')
def list_(show_password: bool):  # we cannot have method 'list' due to conflicts
    """List all available connections.

    pano connection list --show-password
    """
    from panoramic.cli.connections import list_connections_command

    list_connections_command(show_password)


@connection.command(cls=ConnectionAwareCommand)
@click.argument('name', default=None, type=str, required=False)
def test(name: str):
    """Test connections.

    pano connection test
    """
    from panoramic.cli.connections import test_connections_command

    test_connections_command(name)


@cli.group()
def analytics():
    """Analytics subcommand for managing anonymous usage metrics collection preferences."""
    pass


@analytics.command(name='on', cls=ConfigAwareCommand)
def analytics_on():
    """Opt in to anonymous usage analytics.

    pano analytics on
    """
    from panoramic.cli.analytics import opt_in_command

    opt_in_command()


@analytics.command(name='off', cls=ConfigAwareCommand)
def analytics_off():
    """Opt out off anonymous usage analytics.

    pano analytics off
    """
    from panoramic.cli.analytics import opt_out_command

    opt_out_command()


@analytics.command(name='id', cls=ConfigAwareCommand)
def analytics_id():
    """Display current anonymous usage tracking id.

    pano analytics id
    """
    from panoramic.cli.analytics import show_tracking_id_command

    show_tracking_id_command()



import logging

from panoramic.cli.executor import Executor
from panoramic.cli.remote.writer import RemoteWriter
from panoramic.cli.state import Action

logger = logging.getLogger(__name__)


class RemoteExecutor(Executor):

    """Executes actions against remote API."""

    company_name: str
    writer: RemoteWriter

    def __init__(self, company_name: str):
        self.company_name = company_name
        self.writer = RemoteWriter(company_name)

    def _execute(self, action: Action):
        if action.is_creation:
            assert action.desired is not None
            self.writer.write(action.desired)
        elif action.is_deletion:
            assert action.current is not None
            self.writer.delete(action.current)
        else:
            # assume update
            assert action.desired is not None
            self.writer.write(action.desired)



import json
import logging
import platform
import shutil
import time
from datetime import datetime, timedelta
from typing import Any, Dict, List

# Unfortunate naming choice of Segment analytics python lib.
# May be fixed in future, follow: https://github.com/segmentio/analytics-python/issues/154 .
import analytics  # type: ignore

from panoramic.cli.__version__ import __version__
from panoramic.cli.config.analytics import get_write_key
from panoramic.cli.config.analytics import is_enabled as config_is_enabled
from panoramic.cli.config.auth import get_client_id
from panoramic.cli.config.storage import read_config, update_config
from panoramic.cli.file_utils import (
    append_json_line,
    read_json_lines,
    read_yaml,
    truncate_file,
    write_yaml,
)
from panoramic.cli.paths import Paths
from panoramic.cli.print import echo_info

logger = logging.getLogger(__name__)

MINIMAL_FLUSH_DURATION = timedelta(minutes=10)
MINIMAL_FLUSH_EVENTS = 10

"""
Usage: currently only way to track data is by using `write_command_event` after execution of each CLI command.
If we decide to extend analytics with extra information, implement your own method like `write_request_event`
that will track your data, but there is no need to call flush at the end.
Data will be stored to events journal and flushed based on conditions set here.
"""


def write_command_event(name: str, group: str, start_time: float, error: str = '') -> None:
    """
    Generate command execution context and store it in analytics events file as JSON line.
    Truncate blacklisted parameters, like password.
    Try to flush data to external source based on various conditions.
    """
    if not is_enabled():
        return

    properties = {
        'command': {
            'name': ' '.join((group, name)),
            'duration_seconds': time.time() - start_time,
            'error': error,
        },
        'timestamp': time.time(),
        'platform': platform.platform(),
        'app_version': __version__,
        'python_version': platform.python_version(),
        'user_id': _get_user_id(),
        'success': error == '',
    }

    # Store event into temporary JSON lines events file.
    append_json_line(Paths.analytics_events_file(), properties)

    # Do not crash the command because we are unable to connect with external source.
    try:
        _flush()
    except Exception as e:
        logger.debug("Analytics flush failed.", exc_info=e)


def is_enabled() -> bool:
    """Check if user opted out of usage metrics recording."""
    if not config_is_enabled():
        return False

    config = read_config('analytics')
    return config.get('enabled', False)


def opt_in_command() -> None:
    """CLI command to opt in to anonymous usage analytics."""
    update_config('analytics', {'enabled': True})


def opt_out_command() -> None:
    """CLI command to opt out off anonymous usage analytics."""
    update_config('analytics', {'enabled': False})

    if Paths.analytics_dir().is_dir():
        shutil.rmtree(Paths.analytics_dir())


def show_tracking_id_command() -> None:
    """CLI command to display users anonymous tracking id."""
    echo_info(_get_user_id())


def _get_user_id() -> str:
    """Get client_id from configuration step. If doesn't exist yet use just empty string."""
    try:
        return get_client_id()
    except KeyError:
        return ''

    # Old implementation, can be removed later when above solution if verified and released.
    # """Look for unique user ID that is used while storing events.
    # If no user ID was found, generate new id for user and store it in analytics metadata file.
    # """
    # metadata = _read_metadata()
    # if 'user_id' in metadata:
    #     return metadata['user_id']
    #
    # user_id = str(uuid.uuid4())
    # _update_metadata({'user_id': user_id})
    # return user_id


def _read_events() -> List[Dict[str, Any]]:
    """Read all events from JSON lines events file."""
    events_file = Paths.analytics_events_file()
    if not events_file.is_file():
        return []

    return read_json_lines(Paths.analytics_events_file())


def _read_metadata() -> Dict[str, Any]:
    """Read analytics metadata file containing information like last flush time or anonymous unique user ID."""
    metadata_file = Paths.analytics_metadata_file()
    if not metadata_file.is_file():
        return {}

    metadata = read_yaml(metadata_file)
    return metadata


def _update_metadata(data: Dict[str, Any]) -> None:
    """Update analytics metadata file. For example to update last flush time."""
    metadata = _read_metadata()
    metadata.update(data)
    write_yaml(Paths.analytics_metadata_file(), metadata)


def _get_last_flush_time() -> datetime:
    """Datetime of last flush."""
    metadata = _read_metadata()
    if 'last_flush_time' in metadata:
        return datetime.fromtimestamp(metadata['last_flush_time'])
    return datetime.now()


def _flush() -> None:
    """Perform flush to external system.
    This will iterate through all events from current interval and push them to external Analytics system.
    After all events were successfully sent it will truncate events and update last flush time interval.
    """
    proceed = False
    # Check if flush happen more than MINIMAL_FLUSH_DURATION ago.
    if _get_last_flush_time() < datetime.now() - MINIMAL_FLUSH_DURATION:
        proceed = True

    # Check if number of events since last flush if bigger than MINIMAL_FLUSH_EVENTS threshold.
    events = _read_events()
    if len(events) > MINIMAL_FLUSH_EVENTS:
        proceed = True

    if not proceed:
        return

    analytics.write_key = get_write_key()
    if analytics.write_key is None:
        return

    logger.debug("Flushing analytics events.")

    # Because sending of analytics is asynchronous, only way to get error is by adding an on_error callback,
    # As we don't want to delete events that weren't sent we specify analytics_sent_successfully
    # that can be overridden inside this callback to keep the events.
    analytics_sent_successfully = True

    def on_error(*args, **kwargs):
        logger.debug("Failed to send analytics events.", *args, **kwargs)
        nonlocal analytics_sent_successfully
        analytics_sent_successfully = False

    analytics.on_error = on_error
    for event in events:
        analytics.track(_get_user_id(), json.dumps(event))

    # Batching is automatic according to:
    # https://segment.com/docs/connections/sources/catalog/libraries/server/python/#batching
    analytics.flush()

    # Truncate events file after all events were flushed and store time of the flush.
    if analytics_sent_successfully:
        truncate_file(Paths.analytics_events_file())
        _update_metadata({'last_flush_time': time.time()})
